from typing import Optional, List, Union, Dict, Tuple

import pandas as pd
import plotly.graph_objects as go
import plotly.express as px

from elphick.geomet.utils.size import mean_size
from elphick.geomet.utils.viz import plot_parallel


def parallel_plot(data: pd.DataFrame,
                  color: Optional[str] = None,
                  vars_include: Optional[List[str]] = None,
                  vars_exclude: Optional[List[str]] = None,
                  title: Optional[str] = None,
                  include_dims: Optional[Union[bool, List[str]]] = True,
                  plot_interval_edges: bool = False) -> go.Figure:
    """Create an interactive parallel plot

    Useful to explore multidimensional data like mass-composition data

    Args:
        data: The DataFrame to plot
        color: Optional color variable
        vars_include: Optional List of variables to include in the plot
        vars_exclude: Optional List of variables to exclude in the plot
        title: Optional plot title
        include_dims: Optional boolean or list of dimension to include in the plot.  True will show all dims.
        plot_interval_edges: If True, interval edges will be plotted instead of interval mid

    Returns:

    """
    df: pd.DataFrame = data.copy()
    if vars_include is not None:
        missing_vars = set(vars_include).difference(set(df.columns))
        if len(missing_vars) > 0:
            raise KeyError(f'var_subset provided contains variable not found in the data: {missing_vars}')
        df = df[vars_include]
    if vars_exclude:
        df = df[[col for col in df.columns if col not in vars_exclude]]

    if include_dims is True:
        df.reset_index(inplace=True)
    elif isinstance(include_dims, List):
        for d in include_dims:
            df.reset_index(d, inplace=True)

    interval_cols: Dict[str, int] = {col: i for i, col in enumerate(df.columns) if df[col].dtype == 'interval'}

    for col, pos in interval_cols.items():
        if plot_interval_edges:
            df.insert(loc=pos + 1, column=f'{col}_left', value=df[col].array.left)
            df.insert(loc=pos + 2, column=f'{col}_right', value=df[col].array.right)
            df.drop(columns=col, inplace=True)
        else:
            # workaround for https://github.com/Elphick/mass-composition/issues/1
            if col == 'size':
                df[col] = mean_size(pd.arrays.IntervalArray(df[col]))
            else:
                df[col] = df[col].array.mid

    fig = plot_parallel(data=df, color=color, title=title)
    return fig


def comparison_plot(data: pd.DataFrame,
                    x: str, y: str,
                    facet_col_wrap: int = 3,
                    color: Optional[str] = None,
                    trendline: bool = False,
                    trendline_kwargs: Optional[Dict] = None) -> go.Figure:
    """Comparison Plot with multiple x-y scatter plots

    Args:
        data: DataFrame, in tidy (tall) format, with columns for x and y
        x: The x column
        y: The y column
        facet_col_wrap: the number of subplots per row before wrapping
        color: The optional variable to color by. If None color will be by Node
        trendline: If True add trendlines
        trendline_kwargs: Allows customising the trendline: ref: https://plotly.com/python/linear-fits/.  Note: Axis
         scaling across components can be affected if using {'trendline_scope': 'trendline_scope'}.

    Returns:
        plotly Figure
    """
    if trendline:
        if trendline_kwargs is None:
            trendline_kwargs = {'trendline': 'ols'}
        else:
            if 'trendline' not in trendline_kwargs:
                trendline_kwargs['trendline'] = "ols"
    else:
        trendline_kwargs = {'trendline': None}

    data['residual'] = data[x] - data[y]
    fig = px.scatter(data, x=x, y=y, color=color,
                     facet_col='variable', facet_col_wrap=facet_col_wrap,
                     hover_data=['residual'],
                     **trendline_kwargs)

    # fig.print_grid()
    # add y=x based on data per subplot
    variable_order = list(data['variable'].unique())
    d_subplots = subplot_index_by_title(fig, variable_order)

    for k, v in d_subplots.items():
        tmp_df = data.query('variable==@k')
        limits = [min([tmp_df[x].min(), tmp_df[y].min()]),
                  max([tmp_df[x].max(), tmp_df[y].max()])]

        equal_trace = go.Scatter(x=limits, y=limits,
                                 line_color="gray", name="y=x", mode='lines', legendgroup='y=x', showlegend=False)
        fig.add_trace(equal_trace, row=v[0], col=v[1], exclude_empty_subplots=True)
        sp = fig.get_subplot(v[0], v[1])
        fig.update_xaxes(scaleanchor=sp.xaxis.anchor, scaleratio=1, row=v[0], col=v[1])

    fig.update_traces(selector=-1, showlegend=True)
    fig.for_each_yaxis(lambda _y: _y.update(showticklabels=True, matches=None))
    fig.for_each_xaxis(lambda _x: _x.update(showticklabels=True, matches=None))

    return fig


def subplot_index_by_title(fig, variable_order: List[str]) -> Dict['str', Tuple[int, int]]:
    """Map of subplot index by title

    Assumes consistency by plotly between axes numbering and annotation order.

    Args:
        fig: The figure including subplots with unique titles
        variable_order: the variables in order top-left to bottom-right

    Returns:
        Dict keyed by title with tuple of subplot positions
    """

    d_subplots: Dict = {}
    i = 0
    for r in range(len(fig._grid_ref), 0, -1):
        for c in range(1, len(fig._grid_ref[0]) + 1, 1):
            if i < len(variable_order):
                d_subplots[variable_order[i]] = (r, c)
            i += 1

    return d_subplots



import importlib
from functools import partial

import numpy as np
import pandas as pd


def perfect(x: np.ndarray, d50: float) -> np.ndarray:
    """A perfect partition
    
    Args:
        x: The input dimension, e.g. size or density
        d50: The cut-point

    Returns:

    """
    pn: np.ndarray = np.where(x >= d50, 1.0, 0.0)
    return pn


def napier_munn(x: np.ndarray, d50: float, ep: float) -> np.ndarray:
    """The Napier-Munn partition (1998)

    REF: https://www.sciencedirect.com/science/article/pii/S1474667016453036

    Args:
        x: The input dimension, e.g. size or density
        d50: The cut-point
        ep: The Escarte Probable

    Returns:

    """
    pn: np.ndarray = 1 / (1 + np.exp(1.099 * (d50 - x) / ep))
    return pn


def napier_munn_size(size: np.ndarray, d50: float, ep: float) -> np.ndarray:
    return napier_munn(size, d50, ep)


def napier_munn_density(density: np.ndarray, d50: float, ep: float) -> np.ndarray:
    return napier_munn(density, d50, ep)


napier_munn_size_1mm = partial(napier_munn_size, d50=1.0, ep=0.1)


def load_partition_function(module_name, function_name):
    module = importlib.import_module(module_name)
    return getattr(module, function_name)

# if __name__ == '__main__':
#     da = np.arange(0, 10)
#     PN = perfect(da, d50=6.3)
#     df = pd.DataFrame([da, PN], index=['da', 'pn']).T
#     print(df)
#
#     da = np.arange(0, 10)
#     PN = napier_munn(da, d50=6.3, ep=0.1)
#     df = pd.DataFrame([da, PN], index=['da', 'pn']).T
#     print(df)



from typing import Dict

import networkx as nx
import numpy as np
from networkx import DiGraph, multipartite_layout


def digraph_linear_layout(g, orientation: str = "vertical", scale: float = -1.0):
    """Position nodes of a digraph in layers of straight lines.

    Parameters
    ----------
    g : NetworkX graph or list of nodes
        A position will be assigned to every node in G.

    orientation : string (default='vertical')

    scale : number (default: 1)
        Scale factor for positions.


    Returns
    -------
    pos : dict
        A dictionary of positions keyed by node.

    Examples
    --------
    >>> G = nx.complete_multipartite_graph(28, 16, 10)
    >>> pos = digraph_linear_layout(g)

    Notes
    -----
    Intended for use with DiGraphs with a single degree 1 node with an out-edge

    This algorithm currently only works in two dimensions and does not
    try to minimize edge crossings.

    """

    src_nodes = [n for n, d in g.in_degree() if d == 0]
    g.nodes[src_nodes[0]]['_dist'] = 0
    for x_dist in range(1, len(g.nodes) + 1):
        nodes_at_x_dist: dict = nx.descendants_at_distance(g, src_nodes[0], x_dist)
        if not nodes_at_x_dist:
            break
        else:
            for node in nodes_at_x_dist:
                g.nodes[node]['_dist'] = x_dist

    # Ensure all nodes have a _dist attribute
    for node in g.nodes:
        if '_dist' not in g.nodes[node]:
            try:
                g.nodes[node]['_dist'] = nx.shortest_path_length(g, source=src_nodes[0], target=node)
            except nx.NetworkXNoPath:
                g.nodes[node]['_dist'] = 0.0  # or any other default distance

    if orientation == 'vertical':
        orientation = 'horizontal'
    elif orientation == 'horizontal':
        orientation = 'vertical'
        scale = -scale
    else:
        raise ValueError("orientation argument not in 'vertical'|'horizontal'")

    pos = multipartite_layout(g, subset_key="_dist", align=orientation, scale=scale)

    for node in g.nodes:
        g.nodes[node].pop('_dist')

    return pos



import webbrowser
from pathlib import Path
from typing import Dict

import pandas as pd
import platformdirs
import pooch
from pooch import Unzip, Pooch


class Downloader:
    def __init__(self):
        """Instantiate a Downloader
        """

        self.register: pd.DataFrame = pd.read_csv(Path(__file__).parent / 'register.csv', index_col=False)

        self.dataset_hashes: Dict = self.register[['target', 'target_sha256']].set_index('target').to_dict()[
            'target_sha256']

        self.downloader: Pooch = pooch.create(path=Path(platformdirs.user_cache_dir('mass_composition', 'elphick')),
                                              base_url="https://github.com/elphick/mass-composition/raw/main/docs"
                                                       "/source/_static/",
                                              version=None,
                                              version_dev=None,
                                              registry={**self.dataset_hashes})

    def load_data(self, datafile: str = 'size_by_assay.zip', show_report: bool = False) -> pd.DataFrame:
        """
        Load the 231575341_size_by_assay data as a pandas.DataFrame.
        """
        if datafile not in self.dataset_hashes.keys():
            raise KeyError(f"The file {datafile} is not in the registry containing: {self.dataset_hashes.keys()}")

        fnames = self.downloader.fetch(datafile, processor=Unzip())
        if show_report:
            webbrowser.open(str(Path(fnames[0]).with_suffix('.html')))
        data = pd.read_csv(Path(fnames[0]).with_suffix('.csv'))
        return data




from typing import Optional, Iterable, Union

import numpy as np
import pandas as pd
from scipy.interpolate import pchip_interpolate

from elphick.geomet.utils.pandas import composition_to_mass, mass_to_composition, weight_average


def mass_preserving_interp(df_intervals: pd.DataFrame, interval_edges: Union[Iterable, int],
                           include_original_edges: bool = True, precision: Optional[int] = None,
                           mass_wet: Optional[str] = 'mass_wet', mass_dry: str = 'mass_dry',
                           interval_data_as_mass: bool = False) -> pd.DataFrame:
    """Interpolate with zero mass loss using pchip

    This interpolates data_vars independently for a single dimension (coord) at a time.

    The function will:
    - convert from relative composition (%) to absolute (mass) (subject to interval_data_as_mass argument)
    - convert the index from interval to a float representing the right edge of the interval
    - cumsum to provide monotonic increasing data
    - interpolate with a pchip spline to preserve mass
    - diff to recover the original fractional data
    - reconstruct the interval index from the right edges
    - convert from absolute to relative composition

    Args:
        df_intervals: A pd.DataFrame with a single interval index, with mass, composition context.
        interval_edges: The values of the new grid (interval edges).  If an int, will up-sample by that factor, for
         example the value of 10 will automatically define edges that create 10 x the resolution (up-sampled).
        include_original_edges: If True include the original index edges in the result
        precision: Number of decimal places to round the index (edge) values.
        mass_wet: The wet mass column, not optional.  Consider solve_mass_moisture prior to this call if needed.
        mass_dry: The dry mass column, not optional.  Consider solve_mass_moisture prior to this call if needed.
        interval_data_as_mass: If True, the data is assumed to be mass data, not composition data, negating the need
            to convert to mass.

    Returns:

    """

    if not isinstance(df_intervals.index, pd.IntervalIndex):
        raise NotImplementedError(f"The index `{df_intervals.index}` of the dataframe is not a pd.Interval. "
                                  f"Only 1D interval indexes are valid")

    composition_in: pd.DataFrame = df_intervals.copy()

    if isinstance(interval_edges, int):
        grid_vals = _upsample_grid_by_factor(indx=composition_in.sort_index().index, factor=interval_edges)
    else:
        grid_vals = np.sort(np.array(interval_edges))

    if precision is not None:
        composition_in.index = pd.IntervalIndex.from_arrays(np.round(df_intervals.index.left, precision),
                                                            np.round(df_intervals.index.right, precision),
                                                            closed=df_intervals.index.closed,
                                                            name=df_intervals.index.name)

        grid_vals = np.round(grid_vals, precision)

    if include_original_edges:
        original_edges = np.hstack([df_intervals.index.left, df_intervals.index.right])
        grid_vals = np.sort(np.unique(np.hstack([grid_vals, original_edges])))

    if not isinstance(grid_vals, np.ndarray):
        grid_vals = np.array(grid_vals)

    if not interval_data_as_mass:
        # convert from relative composition (%) to absolute (mass)
        mass_in: pd.DataFrame = composition_to_mass(composition_in, mass_wet=mass_wet, mass_dry=mass_dry)
    else:
        mass_in: pd.DataFrame = composition_in.copy()
    # convert the index from interval to a float representing the right edge of the interval
    mass_in.index = mass_in.index.right
    # add a row of zeros
    mass_in = pd.concat(
        [mass_in, pd.Series(0, index=mass_in.columns, name=composition_in.index.left.min()).to_frame().T],
        axis=0).sort_index(ascending=True)
    # cumsum to provide monotonic increasing data
    mass_cum: pd.DataFrame = mass_in.cumsum()
    # if the new grid extrapolates (on the coarse side), mass will be lost, so we assume that when extrapolating.
    # the mass in the extrapolated fractions is zero.  By inserting these records the spline will conform.
    x_extra = grid_vals[grid_vals > mass_cum.index.max()]
    if len(x_extra) > 0:
        cum_max: pd.Series = mass_cum.iloc[-1, :]
        mass_cum = mass_cum.reindex(index=mass_cum.index.append(pd.Index(x_extra)))  # reindex to enable insert
        mass_cum.loc[x_extra, :] = cum_max.values
    #  interpolate with a pchip spline to preserve mass
    chunks = []
    for col in mass_cum:
        tmp = mass_cum[col].dropna()  # drop any missing values
        new_vals = pchip_interpolate(tmp.index.values, tmp.values, grid_vals)
        chunks.append(new_vals)
    mass_cum_upsampled: pd.DataFrame = pd.DataFrame(chunks, index=mass_in.columns, columns=grid_vals).T
    # diff to recover the original fractional data
    mass_fractions_upsampled: pd.DataFrame = mass_cum_upsampled.diff().dropna(axis=0)
    # reconstruct the interval index from the grid
    mass_fractions_upsampled.index = pd.IntervalIndex.from_arrays(left=grid_vals[:-1],
                                                                  right=grid_vals[1:],
                                                                  closed=df_intervals.index.closed,
                                                                  name=df_intervals.index.name)
    interval_spans: dict[pd.Interval, float] = {interval: interval.right - interval.left for interval in
                                                mass_fractions_upsampled.index}
    zero_spans = [k for k, v in interval_spans.items() if v == 0]
    if len(zero_spans) > 1:
        raise ValueError(f"The interpolated index contains zero width intervals on left edges: {zero_spans}")

    # convert from absolute to relative composition
    res = mass_to_composition(mass_fractions_upsampled, mass_wet=mass_wet, mass_dry=mass_dry).sort_index(
        ascending=False)

    # confirm the weight average is preserved
    if not np.isclose(mass_in.sum().sum(), mass_fractions_upsampled.sum().sum(), rtol=1e-6):
        raise ValueError("The mass is not preserved in the interpolation")
    return res


def _upsample_grid_by_factor(indx: pd.IntervalIndex, factor):
    # TODO: must be a better way than this - vectorised?
    grid_vals: list = [indx.left.min()]
    for interval in indx:
        increment = (interval.right - interval.left) / factor
        for i in range(0, factor):
            grid_vals.append(interval.left + (i + 1) * increment)
    grid_vals.sort()
    return grid_vals


def mass_preserving_interp_2d(intervals: pd.DataFrame, interval_edges: dict[str, Iterable],
                              include_original_edges: bool = True, precision: Optional[int] = None,
                              mass_dry: str = 'mass_dry') -> pd.DataFrame:
    """Interpolate 2D interval data with zero mass loss using pchip

    This function applies mass-preserving up-sampling to 2D interval data.  The function will:
    - resample the first dimension using the mass_preserving_interp function
    - resample the second dimension using the mass_preserving_interp function at each of the new first
        dimension intervals
    - apply the upsampled mass proportions of the second dimension to the first dimension to create the final result

    Args:
        intervals: Dataframe with two pd.IntervalIndexes, in mass-composition space.
        interval_edges: Dict of the values of the new grid (interval edges) for each dimension, keyed by
         index name (dimension).
        include_original_edges: If True include the original index edges in the result
        precision: Number of decimal places to round the index (edge) values.
        mass_dry: The dry mass column, not optional.  Consider solve_mass_moisture prior to this call if needed.

    Returns:

    """

    # reduce the dataframe to the first dimension
    dim_1_name, dim_2_name = intervals.index.names[0], intervals.index.names[1]
    dim_1: pd.DataFrame = intervals.groupby(dim_1_name).apply(weight_average, **{'mass_dry': mass_dry})

    # interpolate the first dimension
    dim_1_interp: pd.DataFrame = mass_preserving_interp(dim_1, interval_edges=interval_edges[dim_1_name],
                                                        include_original_edges=include_original_edges,
                                                        precision=precision, mass_dry=mass_dry)

    chunks: list = []
    # iterate the original dim_1 fractions
    for dim_1_interval in dim_1.index:
        # reduce the dataframe to the second dimension for the current dim1 interval
        dim_2: pd.DataFrame = intervals.loc[dim_1_interval].copy()
        # interpolate the second dimension
        dim_2_interp: pd.DataFrame = mass_preserving_interp(dim_2, interval_edges=interval_edges[dim_2_name],
                                                            include_original_edges=include_original_edges,
                                                            precision=precision, mass_dry=mass_dry)
        # convert to recovery to enable proportioning of the interpolated dim_1 values
        dim_2_mass: pd.DataFrame = composition_to_mass(dim_2_interp, mass_dry=mass_dry)
        dim_2_deportment: pd.DataFrame = dim_2_mass.div(dim_2_mass.sum(axis=0), axis=1)

        # Filter the intervals from dim_1_interp that fall within the provided dim_1_interval, convert to mass
        filtered_intervals_mass = composition_to_mass(
            dim_1_interp.loc[dim_1_interp.index.map(lambda x: x.overlaps(dim_1_interval))].copy(), mass_dry=mass_dry)

        # expand the dimensions of the interpolated dim_1 data to include the second upsampled dimension
        for dim_1_interp_interval in filtered_intervals_mass.index:
            # proportion the dim_1 interpolated values by the dim_2 recovery
            new_vals: pd.DataFrame = dim_2_deportment.mul(filtered_intervals_mass.loc[dim_1_interp_interval].values)
            # create a multiindex by combining the dim_1 and dim_2 intervals
            new_index = pd.MultiIndex.from_arrays(
                [pd.IntervalIndex([dim_1_interp_interval] * len(new_vals)), new_vals.index])
            new_vals.index = new_index
            chunks.append(new_vals)

    # concatenate the results
    res: pd.DataFrame = pd.concat(chunks, axis=0)
    res.index.names = [dim_1_name, dim_2_name]
    # convert to composition
    res = mass_to_composition(res, mass_dry=mass_dry)
    return res



import copy
import inspect
import logging
import re
import uuid
from abc import ABC
from pathlib import Path
from typing import Optional, Union, Literal, TypeVar, TYPE_CHECKING

import numpy as np
import pandas as pd
import plotly.express as px
import plotly.graph_objects as go

from elphick.geomet.config import read_yaml
from elphick.geomet.utils.components import get_components, is_compositional
from elphick.geomet.utils.moisture import solve_mass_moisture
from elphick.geomet.utils.pandas import mass_to_composition, composition_to_mass, composition_factors
from elphick.geomet.utils.timer import log_timer
from .config.config_read import get_column_config
from .plot import parallel_plot, comparison_plot

if TYPE_CHECKING:
    from elphick.geomet.flowsheet.stream import Stream

# generic type variable, used for type hinting, to indicate that the type is a subclass of MassComposition
MC = TypeVar('MC', bound='MassComposition')


def filter_kwargs(cls, **kwargs):
    valid_params = inspect.signature(cls.__init__).parameters
    res = {k: v for k, v in kwargs.items() if k in valid_params}
    return res


class MassComposition(ABC):
    def __init__(self,
                 data: Optional[pd.DataFrame] = None,
                 name: Optional[str] = None,
                 moisture_in_scope: bool = True,
                 mass_wet_var: Optional[str] = None,
                 mass_dry_var: Optional[str] = None,
                 moisture_var: Optional[str] = None,
                 component_vars: Optional[list[str]] = None,
                 composition_units: Literal['%', 'ppm', 'ppb'] = '%',
                 components_as_symbols: bool = True,
                 ranges: Optional[dict[str, list]] = None,
                 config_file: Optional[Path] = None):
        """

        Args:
            data: The input data
            name: The name of the sample
            moisture_in_scope: Whether the moisture is in scope.  If False, only dry mass is processed.
            mass_wet_var: The name of the wet mass column
            mass_dry_var: The name of the dry mass column
            moisture_var: The name of the moisture column
            component_vars: The names of the chemical columns
            components_as_symbols: If True, convert the composition variables to symbols, e.g. Fe
            ranges: The range of valid data for each column in the data
            config_file: The configuration file
        """

        self._logger = logging.getLogger(name=self.__class__.__name__)

        if config_file is None:
            config_file = Path(__file__).parent / './config/mc_config.yml'
        self.config = read_yaml(config_file)

        self.name: str = name
        self.moisture_in_scope: bool = moisture_in_scope
        self.mass_wet_var: Optional[str] = mass_wet_var
        self.mass_dry_var: str = mass_dry_var
        self.moisture_var: Optional[str] = moisture_var
        self.component_vars: Optional[list[str]] = component_vars  # TODO: check if this is redundant and remove.
        self.composition_units: Literal['%', 'ppm', 'ppb'] = composition_units
        self.composition_factor: int = composition_factors[composition_units]
        self.components_as_symbols: bool = components_as_symbols

        self._mass_data: Optional[pd.DataFrame] = None
        self._supplementary_data = None
        self._aggregate = None

        # set the data
        self.data = data

        # add the OOR status object
        self.status = OutOfRangeStatus(self, ranges)

    @property
    @log_timer
    def data(self) -> Optional[pd.DataFrame]:
        if self._mass_data is not None:
            # convert chem mass to composition
            mass_comp_data = mass_to_composition(self._mass_data,
                                                 mass_wet=self.mass_wet_var, mass_dry=self.mass_dry_var,
                                                 moisture_column_name='H2O' if self.components_as_symbols else (
                                                     self.moisture_var if self.moisture_var is not None else 'h2o'),
                                                 component_columns=self.composition_columns,
                                                 composition_units=self.composition_units)

            # append the supplementary vars
            return pd.concat([mass_comp_data, self._supplementary_data], axis=1)
        return None

    @data.setter
    @log_timer
    def data(self, value):
        if value is not None:
            # Convert column names to symbols if components_as_symbols is True
            if self.components_as_symbols:
                symbol_dict = is_compositional(value.columns, strict=False)
                value.columns = [symbol_dict.get(col, col) for col in value.columns]

            # the config provides regex search keys to detect mass and moisture columns if they are not specified.
            mass_totals = self._solve_mass(value)
            composition, supplementary_data = self._get_non_mass_data(value)

            self._supplementary_data = supplementary_data

            self.mass_data = composition_to_mass(pd.concat([mass_totals, composition], axis=1),
                                                 mass_wet=self.mass_wet_var, mass_dry=self.mass_dry_var,
                                                 moisture_column_name=self.moisture_column,
                                                 component_columns=composition.columns,
                                                 composition_units=self.composition_units,
                                                 return_moisture=False)
            self._logger.debug(f"Data has been set.")

        else:
            self._mass_data = None

    @property
    def mass_data(self):
        return self._mass_data

    @mass_data.setter
    def mass_data(self, value):
        self._mass_data = value
        # Recalculate the aggregate whenever the data changes
        self.aggregate = self.weight_average()

    def get_mass_data(self, include_moisture: bool = True) -> pd.DataFrame:
        """Get the mass data

        Args:
            include_moisture: If True (and moisture is in scope), include the moisture mass column

        Returns:

        """
        if include_moisture and self.moisture_in_scope:
            moisture_mass = self._mass_data[self.mass_wet_var] - self._mass_data[self.mass_dry_var]
            mass_data: pd.DataFrame = self._mass_data.copy()
            mass_data.insert(loc=2, column=self.moisture_column, value=moisture_mass)
            return mass_data
        return self._mass_data

    @property
    def aggregate(self) -> pd.DataFrame:
        if self._aggregate is None and self._mass_data is not None:
            self._aggregate = self.weight_average()
        return self._aggregate

    @aggregate.setter
    def aggregate(self, value):
        self._aggregate = value

    @property
    def variable_map(self) -> Optional[dict[str, str]]:
        """A map from lower case standard names to the actual column names"""
        if self._mass_data is not None:
            existing_columns = list(self._mass_data.columns)
            res = {}
            if self.moisture_in_scope and self.mass_wet_var in existing_columns:
                res['mass_wet'] = self.mass_wet_var
            if self.mass_dry_var in existing_columns:
                res['mass_dry'] = self.mass_dry_var
            if self.moisture_in_scope:
                res['moisture'] = self.moisture_var
                if self.components_as_symbols:
                    res['moisture'] = is_compositional([self.moisture_var], strict=False).get(self.moisture_var,
                                                                                              self.moisture_var)
            if self.composition_columns:
                for col in self.composition_columns:
                    res[col.lower()] = col
            return res
        return None

    @property
    def mass_columns(self) -> Optional[list[str]]:
        if self._mass_data is not None:
            existing_columns = list(self._mass_data.columns)
            res = []
            if self.moisture_in_scope and self.mass_wet_var in existing_columns:
                res.append(self.mass_wet_var)
            if self.mass_dry_var in existing_columns:
                res.append(self.mass_dry_var)
            return res
        return None

    @property
    def moisture_column(self) -> Optional[str]:
        res = 'h2o'
        if self.moisture_in_scope:
            res = self.moisture_var
            if self.components_as_symbols:
                res = is_compositional([res], strict=False).get(res, res)
        return res

    @property
    def composition_columns(self) -> Optional[list[str]]:
        res = None
        if self._mass_data is not None:
            if self.moisture_in_scope:
                res = list(self._mass_data.columns)[2:]
            else:
                res = list(self._mass_data.columns)[1:]
        return res

    @property
    def supplementary_columns(self) -> Optional[list[str]]:
        res = None
        if self._supplementary_data is not None:
            res = list(self._supplementary_data.columns)
        return res

    @property
    def data_columns(self) -> list[str]:
        return [col for col in
                (self.mass_columns + [self.moisture_column] + self.composition_columns + self.supplementary_columns) if
                col is not None]

    def balance_composition(self, epsilon: float = 1.0e-6) -> MC:
        """Balance the composition data

        For records where the component mass exceeds the dry mass, the component masses are reduced proportionally
        to equal the dry mass.  Records where the component mass is less than the dry mass are left unchanged.

        epsilon: A small float to avoid component sums marginally over 100.0

        Returns:
            The object with balanced composition.
        """
        if self._mass_data is not None:
            # calculate the ratio of the sum of the components to the dry mass
            ratio = self._mass_data[self.composition_columns].sum(axis=1) / self._mass_data[self.mass_dry_var]
            if ratio.max() <= 1.0:
                return self

            before_reduction = self._mass_data[self.composition_columns].copy()
            # add a small value to the ratio to avoid component sums marginally over 100.0
            ratio[ratio > 1.0] = ratio[ratio > 1.0] + epsilon
            # to avoid reducing compliant records, clip the ratio at the lower side to 1.0
            ratio = ratio.clip(lower=1.0)
            # apply the ratio to the components
            self._mass_data[self.composition_columns] = self._mass_data[self.composition_columns].div(ratio, axis=0)

            # manage the reporting
            affected_indexes = set(
                self._mass_data.index[np.any(before_reduction != self._mass_data[self.composition_columns], axis=1)])
            # log the action, including the first 50 indexes affected
            affected_indexes_list = sorted(affected_indexes)[:50]

            self.aggregate = self.weight_average()
            self.status = OutOfRangeStatus(self, self.status.ranges)

            self._logger.info(f"Object {self.name} has been balanced. "
                              f"{len(affected_indexes)} records where composition has been reduced "
                              f"to conform to dry mass. "
                              f"Affected indexes (first 50): {affected_indexes_list}")
        return self

    def clip_recovery(self, other: MC, recovery_bounds: tuple[float, float] = (0.01, 0.99),
                      allow_moisture_coercion: bool = True) -> MC:
        """Clip the recovery to the specified bounds and recalculate the estimate.

        Args:
            other: The other MassComposition object, from which the recovery of self is calculated.
            recovery_bounds: The bounds for the recovery between 0.0 and 1.0
            allow_moisture_coercion: if True, allow the wet mass to be modified to maintain the moisture (in the
            case that dry mass is clipped to manage recovery)

        Returns:
            The MassComposition object with the recovery clipped to the bounds.
        """
        recovery: pd.DataFrame = (self.get_mass_data(include_moisture=False) /
                                  other.get_mass_data(include_moisture=False))

        # Limit the recovery to the bounds
        before_clip = recovery.copy()
        recovery = recovery.clip(lower=recovery_bounds[0], upper=recovery_bounds[1]).fillna(0.0)

        # Check if any records were affected
        affected_indexes = set(recovery.index[np.any(before_clip != recovery, axis=1)])
        if affected_indexes:
            # Recalculate the estimate from the bound recovery
            new_mass: pd.DataFrame = recovery * other.get_mass_data(include_moisture=False)[recovery.columns]

            if self.moisture_in_scope and allow_moisture_coercion:
                # Calculate the moisture from the new mass
                new_mass[self.mass_wet_var] = solve_mass_moisture(mass_dry=new_mass[self.mass_dry_var],
                                                                  moisture=self.data[self.moisture_column])

            # Log the top 50 records affected by the recovery coercion
            affected_indexes_list = sorted(affected_indexes)[:50]
            self._logger.info(f"Recovery coercion affected {len(affected_indexes)} records. "
                              f"Affected indexes (first 50): {affected_indexes_list}")

            # Update the mass data of self
            self.update_mass_data(new_mass)
        else:
            self._logger.info("Recovery coercion did not affect any records.")

        return self

    def set_moisture(self, moisture: Union[pd.Series, float, int], mass_to_adjust: Literal['wet', 'dry'] = 'wet') -> MC:
        """Set the moisture to the specified value

        A convenience method for an mc object that modifies the concrete mass to deliver the specified moisture.

        Args:
            moisture: The moisture value to set.  Can be a constant or series.
            mass_to_adjust: The mass to adjust, either 'wet' or 'dry'.

        Returns:

        """

        if not self.moisture_in_scope:
            raise AssertionError("This method is not applicable unless moisture_in_scope property is True.")

        if isinstance(moisture, float) or isinstance(moisture, int):
            # create a series with the same index as the mass data
            moisture = pd.Series(float(moisture), index=self._mass_data.index)
        elif not isinstance(moisture, pd.Series):
            raise TypeError(f"moisture must be a float or a pd.Series, not {type(moisture)}")

        if mass_to_adjust == 'wet':
            self._mass_data[self.mass_wet_var] = solve_mass_moisture(mass_dry=self._mass_data[self.mass_dry_var],
                                                                     moisture=moisture)
        elif mass_to_adjust == 'dry':
            self._mass_data[self.mass_dry_var] = solve_mass_moisture(mass_wet=self._mass_data[self.mass_wet_var],
                                                                     moisture=moisture)
        else:
            raise ValueError(f"mass_to_adjust must be 'wet' or 'dry', not {mass_to_adjust}")

        self.aggregate = self.weight_average()
        self.status = OutOfRangeStatus(self, self.status.ranges)

        return self

    def clip_composition(self, ranges: Optional[dict[str, list[float]]] = None, epsilon: float = 1.0e-05) -> MC:
        """Clip the components

        Clip to the components to within the range provided or the default range for each component.
        This method does not clip moisture - see set_moisture and solve_moisture for that.

        Args:
            ranges: An optional dict defining a list of [lo, hi] floats for each component.  If not provided,
            the default range from the config file will be used.
            epsilon: A small float to ensure the clipped values lie marginally inside the specified range.

        Returns:
            The object with clipped composition.
        """

        # load the default ranges from the config file
        component_ranges: dict = self._get_component_ranges(ranges)

        # define a small value to ensure the clipped values lie marginally inside the specified range.
        # clip the components
        affected_indexes = set()
        for component, component_range in component_ranges.items():
            before_clip = self._mass_data[component].copy()
            # add a small value to the range to ensure the clipped values lie marginally inside the specified range.
            component_range = [component_range[0] + epsilon, component_range[1] - epsilon]
            # define the component mass that aligns with the lower and upper bounds
            component_mass_limits = self._mass_data[self.mass_dry_var].values[:, np.newaxis] * np.array(
                component_range) / self.composition_factor
            # apply the clip to the mass data
            self._mass_data[component] = self._mass_data[component].clip(lower=component_mass_limits[:, 0],
                                                                         upper=component_mass_limits[:, 1])
            affected_indexes.update(self._mass_data.index[before_clip != self._mass_data[component]])

        # log the action, including the first 50 indexes affected
        affected_indexes_list = sorted(affected_indexes)[:50]

        self.aggregate = self.weight_average()
        self.status = OutOfRangeStatus(self, self.status.ranges)

        if len(affected_indexes) > 0:
            self._logger.info(f"Object {self.name} has been clipped. "
                              f"{len(affected_indexes)} records where composition has been clipped to the "
                              f"range: {component_ranges}. "
                              f"Affected indexes (first 50): {affected_indexes_list}")

        return self

    def plot_parallel(self, color: Optional[str] = None,
                      vars_include: Optional[list[str]] = None,
                      vars_exclude: Optional[list[str]] = None,
                      title: Optional[str] = None,
                      include_dims: Optional[Union[bool, list[str]]] = True,
                      plot_interval_edges: bool = False) -> go.Figure:
        """Create an interactive parallel plot

        Useful to explore multidimensional data like mass-composition data

        Args:
            color: Optional color variable
            vars_include: Optional list of variables to include in the plot
            vars_exclude: Optional list of variables to exclude in the plot
            title: Optional plot title
            include_dims: Optional boolean or list of dimension to include in the plot.  True will show all dims.
            plot_interval_edges: If True, interval edges will be plotted instead of interval mid

        Returns:

        """

        if not title and hasattr(self, 'name'):
            title = self.name

        fig = parallel_plot(data=self.data, color=color, vars_include=vars_include, vars_exclude=vars_exclude,
                            title=title,
                            include_dims=include_dims, plot_interval_edges=plot_interval_edges)
        return fig

    def plot_comparison(self, other: MC,
                        color: Optional[str] = None,
                        vars_include: Optional[list[str]] = None,
                        vars_exclude: Optional[list[str]] = None,
                        facet_col_wrap: int = 3,
                        trendline: bool = False,
                        trendline_kwargs: Optional[dict] = None,
                        title: Optional[str] = None) -> go.Figure:
        """Create an interactive parallel plot

        Useful to compare the difference in component values between two objects.

        Args:
            other: the object to compare with self.
            color: Optional color variable
            vars_include: Optional List of variables to include in the plot
            vars_exclude: Optional List of variables to exclude in the plot
            trendline: If True and trendlines
            trendline_kwargs: Allows customising the trendline: ref: https://plotly.com/python/linear-fits/
            title: Optional plot title
            facet_col_wrap: The number of subplot columns per row.

        Returns:

        """
        df_self: pd.DataFrame = self.data.to_dataframe()
        df_other: pd.DataFrame = other.data.to_dataframe()

        if vars_include is not None:
            missing_vars = set(vars_include).difference(set(df_self.columns))
            if len(missing_vars) > 0:
                raise KeyError(f'var_subset provided contains variable not found in the data: {missing_vars}')
            df_self = df_self[vars_include]
        if vars_exclude:
            df_self = df_self[[col for col in df_self.columns if col not in vars_exclude]]
        df_other = df_other[df_self.columns]
        # Supplementary variables are the same for each stream and so will be unstacked.
        supp_cols: list[str] = self.supplementary_columns
        if supp_cols:
            df_self.set_index(supp_cols, append=True, inplace=True)
            df_other.set_index(supp_cols, append=True, inplace=True)

        index_names = list(df_self.index.names)
        cols = list(df_self.columns).copy()

        df_self = df_self[cols].assign(name=self.name).reset_index().melt(id_vars=index_names + ['name'])
        df_other = df_other[cols].assign(name=other.name).reset_index().melt(id_vars=index_names + ['name'])

        df_plot: pd.DataFrame = pd.concat([df_self, df_other])
        df_plot = df_plot.set_index(index_names + ['name', 'variable'], drop=True).unstack(['name'])
        df_plot.columns = df_plot.columns.droplevel(0)
        df_plot.reset_index(level=list(np.arange(-1, -len(index_names) - 1, -1)), inplace=True)

        # set variables back to standard order
        variable_order: dict = {col: i for i, col in enumerate(cols)}
        df_plot = df_plot.sort_values(by=['variable'], key=lambda x: x.map(variable_order))

        fig: go.Figure = comparison_plot(data=df_plot, x=self.name, y=other.name, facet_col_wrap=facet_col_wrap,
                                         color=color, trendline=trendline, trendline_kwargs=trendline_kwargs)
        fig.update_layout(title=title)
        return fig

    def plot_ternary(self, variables: list[str], color: Optional[str] = None,
                     title: Optional[str] = None) -> go.Figure:
        """Plot a ternary diagram

            variables: List of 3 components to plot
            color: Optional color variable
            title: Optional plot title

        """

        df = self.data
        vars_missing: list[str] = [v for v in variables if v not in df.columns]
        if vars_missing:
            raise KeyError(f'Variable/s not found in the dataset: {vars_missing}')

        cols: list[str] = variables
        if color is not None:
            cols.append(color)

        if color:
            fig = px.scatter_ternary(df[cols], a=variables[0], b=variables[1], c=variables[2], color=color)
        else:
            fig = px.scatter_ternary(df[cols], a=variables[0], b=variables[1], c=variables[2])

        if not title and hasattr(self, 'name'):
            title = self.name

        fig.update_layout(title=title)

        return fig

    def weight_average(self, group_by: Optional[str] = None) -> pd.DataFrame:

        if group_by is None:
            composition: pd.DataFrame = pd.DataFrame(
                self._mass_data[self.composition_columns].sum(axis=0) / self._mass_data[
                    self.mass_dry_var].sum() * self.composition_factor).T

            mass_sum = pd.DataFrame(self._mass_data[self.mass_columns].sum(axis=0)).T

            # Recalculate the moisture
            if self.moisture_in_scope:
                mass_sum[self.moisture_column] = solve_mass_moisture(mass_wet=mass_sum[self.mass_columns[0]],
                                                                     mass_dry=mass_sum[self.mass_columns[1]])

            # Create a DataFrame from the weighted averages
            weighted_averages_df = pd.concat([mass_sum, composition], axis=1)

        else:  # group by a variable
            group_var: pd.Series = self._supplementary_data[group_by]
            weighted_averages_df = self._mass_data.groupby(group_var).apply(
                lambda x: pd.DataFrame(
                    x[self.composition_columns].sum(axis=0) / x[self.mass_dry_var].sum() * self.composition_factor).T)
            weighted_averages_df.index = weighted_averages_df.index.droplevel(-1)
            mass_sum = self._mass_data[self.mass_columns].groupby(group_var).sum()
            weighted_averages_df = pd.concat([mass_sum, weighted_averages_df], axis=1)
            if self.moisture_in_scope:
                weighted_averages_df.insert(loc=2, column=self.moisture_column, value=solve_mass_moisture(
                    mass_wet=mass_sum[self.mass_columns[0]],
                    mass_dry=mass_sum[self.mass_columns[1]]))

        return weighted_averages_df

    def _solve_mass(self, value) -> pd.DataFrame:
        """Solve mass_wet and mass_dry from the provided columns.

        Args:
            value: The input data with the column-names provided by the user\

        Returns: The mass data, with the columns mass_wet and mass_dry.  Only mass_dry if moisture_in_scope is False.
        """
        # Auto-detect columns if they are not provided
        mass_dry, mass_wet, moisture = self._extract_mass_moisture_columns(value)

        if mass_dry is None:
            if mass_wet is not None and moisture is not None:
                value[self.mass_dry_var] = solve_mass_moisture(mass_wet=mass_wet, moisture=moisture)
            else:
                msg = (f"mass_dry_var is not provided and cannot be calculated from mass_wet_var and moisture_var "
                       f"for {self.name}")
                self._logger.error(msg)
                raise ValueError(msg)

        if self.moisture_in_scope:
            if mass_wet is None:
                if mass_dry is not None and moisture is not None:
                    value[self.mass_wet_var] = solve_mass_moisture(mass_dry=mass_dry, moisture=moisture)
                else:
                    msg = (
                        f"mass_wet_var is not provided and cannot be calculated from mass_dry_var and moisture_var.  "
                        f"Consider specifying the mass_wet_var, mass_dry_var and moisture_var, or alternatively set "
                        f"moisture_in_scope to False for {self.name}")
                    self._logger.error(msg)
                    raise ValueError(msg)

            if moisture is None:
                if mass_wet is not None and mass_dry is not None:
                    value[self.moisture_var] = solve_mass_moisture(mass_wet=mass_wet, mass_dry=mass_dry)
                else:
                    msg = f"moisture_var is not provided and cannot be calculated from mass_wet_var and mass_dry_var."
                    self._logger.error(msg)
                    raise ValueError(msg)

            mass_totals: pd.DataFrame = value[[self.mass_wet_var, self.mass_dry_var]]
        else:
            mass_totals: pd.DataFrame = value[[self.mass_dry_var]]

        return mass_totals

        # Helper method to extract column

    def _extract_column(self, value, var_type):
        var = getattr(self, f"{var_type}_var")
        if var is None:
            var = next((col for col in value.columns if
                        re.search(self.config['vars'][var_type]['search_regex'], col,
                                  re.IGNORECASE)), self.config['vars'][var_type]['default_name'])
        return var

    def _extract_mass_moisture_columns(self, value):
        if self.mass_wet_var is None:
            self.mass_wet_var = self._extract_column(value, 'mass_wet')
        if self.mass_dry_var is None:
            self.mass_dry_var = self._extract_column(value, 'mass_dry')
        if self.moisture_var is None:
            self.moisture_var = self._extract_column(value, 'moisture')
        mass_wet = value.get(self.mass_wet_var)
        mass_dry = value.get(self.mass_dry_var)
        moisture = value.get(self.moisture_var)
        return mass_dry, mass_wet, moisture

    def _get_non_mass_data(self, value: Optional[pd.DataFrame]) -> (Optional[pd.DataFrame], Optional[pd.DataFrame]):
        """
        Get the composition data and supplementary data.  Extract only the composition columns specified,
        otherwise detect the compositional columns
        """
        composition = None
        supplementary = None
        if value is not None:
            if self.component_vars is None:
                non_mass_cols: list[str] = [col for col in value.columns if
                                            col not in [self.mass_wet_var, self.mass_dry_var, self.moisture_var, 'h2o',
                                                        'H2O', 'H2O']]
                component_cols: list[str] = get_components(value[non_mass_cols].columns, strict=False)
            else:
                component_cols: list[str] = self.component_vars
            composition = value[component_cols]

            supplementary_cols: list[str] = [col for col in value.columns if
                                             col not in component_cols + [self.mass_wet_var, self.mass_dry_var,
                                                                          self.moisture_var, 'h2o',
                                                                          'H2O', 'H2O']]
            supplementary = value[supplementary_cols]

        return composition, supplementary

    def __deepcopy__(self, memo):
        # Create a new instance of our class
        new_obj = self.__class__()
        memo[id(self)] = new_obj

        # Copy each attribute
        for attr, value in self.__dict__.items():
            setattr(new_obj, attr, copy.deepcopy(value, memo))

        return new_obj

    def update_mass_data(self, value: pd.DataFrame):
        if self._mass_data is not None:
            self._mass_data = value
        if self._supplementary_data is not None:
            if self._supplementary_data.index.names != self._mass_data.index.names:  # if indexes have been dropped
                self._supplementary_data.index = self._mass_data.index
            self._supplementary_data = self._supplementary_data.loc[value.index]
        self.aggregate = self.weight_average()
        self.status = OutOfRangeStatus(self, self.status.ranges)

    def filter_by_index(self, index: pd.Index):
        """Update the data by index"""
        if self._mass_data is not None:
            self._mass_data = self._mass_data.loc[index]
        if self._supplementary_data is not None:
            self._supplementary_data = self._supplementary_data.loc[index]
        self.aggregate = self.weight_average()

    def split(self,
              fraction: float,
              name_1: Optional[str] = None,
              name_2: Optional[str] = None,
              include_supplementary_data: bool = False) -> tuple['Stream', 'Stream']:
        """Split the object by mass

        A simple mass split maintaining the same composition

        Args:
            fraction: A constant in the range [0.0, 1.0]
            name_1: The name of the reference object created by the split
            name_2: The name of the complement object created by the split
            include_supplementary_data: Whether to inherit the supplementary variables

        Returns:
            tuple of two objects, the first with the mass fraction specified, the other the complement
        """

        # create_congruent_objects to preserve properties like constraints

        name_1 = name_1 if name_1 is not None else f"{self.name}_1"
        name_2 = name_2 if name_2 is not None else f"{self.name}_2"

        ref: MassComposition = self.create_congruent_object(name=name_1, include_mc_data=True,
                                                            include_supp_data=include_supplementary_data)
        ref.update_mass_data(self._mass_data * fraction)

        comp: MassComposition = self.create_congruent_object(name=name_2, include_mc_data=True,
                                                             include_supp_data=include_supplementary_data)
        comp.update_mass_data(self._mass_data * (1 - fraction))

        # Ensure self and other are Stream objects
        self._convert_to_stream(self)
        self._convert_to_stream(ref)
        self._convert_to_stream(comp)
        self: 'Stream'
        ref: 'Stream'
        comp: 'Stream'

        # create the relationships
        ref.nodes = [self.nodes[1], uuid.uuid4()]
        comp.nodes = [self.nodes[1], uuid.uuid4()]

        return ref, comp

    def add(self, other: MC, name: Optional[str] = None,
            include_supplementary_data: bool = False) -> 'Stream':
        """Add two objects together

        Args:
            other: The other object
            name: The name of the new object
            include_supplementary_data: Whether to include the supplementary data

        Returns:
            The new object
        """

        res: MC = self.create_congruent_object(name=name, include_mc_data=True,
                                               include_supp_data=include_supplementary_data)

        if set(self._mass_data.columns) != set(other._mass_data.columns):
            raise ValueError(f"Mass data columns do not match: {set(self._mass_data.columns)} != "
                             f"{set(other._mass_data.columns)}")
        res.update_mass_data(self._mass_data + other._mass_data)

        # Ensure self and other are Stream objects
        self: 'Stream' = self.to_stream()
        other: 'Stream' = self._convert_to_stream(other)
        res: 'Stream' = self._convert_to_stream(res)

        # create the relationships
        other.nodes = [other.nodes[0], self.nodes[1]]
        res.nodes = [self.nodes[1], uuid.uuid4()]

        return res

    def sub(self, other: MC, name: Optional[str] = None,
            include_supplementary_data: bool = False) -> 'Stream':
        """Subtract other from self

        Args:
            other: The other object
            name: The name of the new object
            include_supplementary_data: Whether to include the supplementary data

        Returns:
            The new object
        """
        res = self.create_congruent_object(name=name, include_mc_data=True,
                                           include_supp_data=include_supplementary_data)

        if set(self._mass_data.columns) != set(other._mass_data.columns):
            raise ValueError(f"Mass data columns do not match: {set(self._mass_data.columns)} != "
                             f"{set(other._mass_data.columns)}")

        res.update_mass_data(self._mass_data - other._mass_data)

        # Ensure self and other are Stream objects
        self._convert_to_stream(self)
        self._convert_to_stream(other)
        self: 'Stream'
        other: 'Stream'

        # create the relationships
        other.nodes = [self.nodes[1], other.nodes[1]]
        res.nodes = [self.nodes[1], uuid.uuid4()]

        return res

    def div(self, other: MC, name: Optional[str] = None,
            include_supplementary_data: bool = False) -> MC:
        """Divide two objects

        Divides self by other, with optional name of the returned object
        Args:
            other: the denominator (or reference) object
            name: name of the returned object
            include_supplementary_data: Whether to include the supplementary data

        Returns:

        """
        new_obj = self.create_congruent_object(name=name, include_mc_data=True,
                                               include_supp_data=include_supplementary_data)
        new_obj.update_mass_data(self._mass_data / other._mass_data)
        return new_obj

    def __str__(self):
        return f"{self.__class__.__name__}: {self.name}\n{self.aggregate.to_dict()}"

    def create_congruent_object(self, name: str,
                                include_mc_data: bool = False,
                                include_supp_data: bool = False) -> MC:
        """Create an object with the same attributes"""
        # Create a new instance of our class
        new_obj = self.__class__()

        # Copy each attribute
        for attr, value in self.__dict__.items():
            if attr == '_mass_data' and not include_mc_data:
                continue
            if attr == '_supplementary_data' and not include_supp_data:
                continue
            setattr(new_obj, attr, copy.deepcopy(value))
        new_obj.name = name
        return new_obj

    def __add__(self, other: MC) -> 'Stream':
        """Add two objects

        Perform the addition with the mass-composition variables only and then append any attribute variables.
        Presently ignores any attribute vars in other
        Args:
            other: object to add to self

        Returns:

        """
        return self.add(other, include_supplementary_data=True)

    def __sub__(self, other: MC) -> 'Stream':
        """Subtract the supplied object from self

        Perform the subtraction with the mass-composition variables only and then append any attribute variables.
        Args:
            other: object to subtract from self

        Returns:

        """
        return self.sub(other, include_supplementary_data=True)

    def to_stream(self) -> 'Stream':
        from elphick.geomet.flowsheet.stream import Stream  # Local import to avoid circular dependency
        if not isinstance(self, Stream):
            self.__class__ = type(self.__class__.__name__, (self.__class__, Stream), {})
            filtered_kwargs = filter_kwargs(self.__class__, **self.__dict__)
            filtered_kwargs['data'] = self.data
            Stream.__init__(self, **filtered_kwargs)  # Initialize Stream properties
        return self

    @staticmethod
    def _convert_to_stream(obj) -> 'Stream':
        from elphick.geomet.flowsheet.stream import Stream  # Local import to avoid circular dependency
        if not isinstance(obj, Stream):
            obj.__class__ = type(obj.__class__.__name__, (obj.__class__, Stream), {})
            filtered_kwargs = filter_kwargs(obj.__class__, **obj.__dict__)
            filtered_kwargs['data'] = obj.data
            Stream.__init__(obj, **filtered_kwargs)  # Initialize Stream properties
        return obj

    def __truediv__(self, other: MC) -> MC:
        """Divide self by the supplied object

        Perform the division with the mass-composition variables only and then append any attribute variables.
        Args:
            other: denominator object, self will be divided by this object

        Returns:

        """

        return self.div(other, include_supplementary_data=True)

    def __eq__(self, other):
        if isinstance(other, MassComposition):
            return self.__dict__ == other.__dict__
        return False

    @classmethod
    def from_mass_dataframe(cls, mass_df: pd.DataFrame,
                            mass_wet: Optional[str] = 'mass_wet',
                            mass_dry: str = 'mass_dry',
                            moisture_column_name: Optional[str] = None,
                            component_columns: Optional[list[str]] = None,
                            composition_units: Literal['%', 'ppm', 'ppb'] = '%',
                            **kwargs) -> MC:
        """
        Class method to create a MassComposition object from a mass dataframe.

        Args:
            mass_df: DataFrame with mass data.
            **kwargs: Additional arguments to pass to the MassComposition constructor.

        Returns:
            A new MassComposition object.
        """
        # Convert mass to composition using the function from the pandas module
        composition_df = mass_to_composition(mass_df, mass_wet=mass_wet, mass_dry=mass_dry,
                                             moisture_column_name=moisture_column_name,
                                             component_columns=component_columns,
                                             composition_units=composition_units)

        # Create a new instance of the MassComposition class
        return cls(data=composition_df, **kwargs)

    def query(self, expr: str, name: Optional[str] = None) -> MC:
        """Reduce the data by a query expression

        Args:
            expr: A pandas query expression
            name: name of the new object

        Returns:
            A new object with the reduced data

        """
        name = name if name is not None else self.name
        res = self.create_congruent_object(name=f"{name} ({expr})", include_mc_data=True,
                                           include_supp_data=True)
        filtered_index = self.data.query(expr).index
        res.update_mass_data(self._mass_data.loc[filtered_index])
        if res.supplementary_columns is not None:
            res._supplementary_data = self._supplementary_data.loc[filtered_index]

        return res

    def compare(self, other: 'MassComposition', comparisons: Union[str, list[str]] = 'recovery',
                explicit_names: bool = True) -> pd.DataFrame:

        comparisons = [comparisons] if isinstance(comparisons, str) else comparisons
        valid_comparisons: set = {'recovery', 'difference', 'divide', 'all'}

        cols = [col for col in self.data.data_vars if col not in self.data.mc.mc_vars_attrs]

        chunks: list[pd.DataFrame] = []
        if 'recovery' in comparisons or comparisons == ['all']:
            df: pd.DataFrame = self._mass_data[self.component_vars] / other._mass_data[self.component_vars]
            if explicit_names:
                df.columns = [f"{self.name}_{col}_{self.config['comparisons']['recovery']}_{other.name}" for col in
                              df.columns]
            chunks.append(df)
        if 'difference' in comparisons or comparisons == ['all']:
            df: pd.DataFrame = self.data[cols] - other.data[cols]
            if explicit_names:
                df.columns = [f"{self.name}_{col}_{self.config['comparisons']['difference']}_{other.name}" for col in
                              df.columns]
            chunks.append(df)
        if 'divide' in comparisons or comparisons == ['all']:
            df: pd.DataFrame = self.data[cols] / other.data[cols]
            if explicit_names:
                df.columns = [f"{self.name}_{col}_{self.config['comparisons']['divide']}_{other.name}" for col in
                              df.columns]
            chunks.append(df)

        if not chunks:
            raise ValueError(f"The comparison argument is not valid: {valid_comparisons}")

        res: pd.DataFrame = pd.concat(chunks, axis=1)

        return res

    def reset_index(self, index_name: str) -> MC:
        res = self.create_congruent_object(name=f"{self.name} (reset_index)", include_mc_data=True,
                                           include_supp_data=True)
        res.update_mass_data(self._mass_data.reset_index(level=index_name, drop=True))
        if res.supplementary_columns is not None:
            res._supplementary_data = self._supplementary_data.reset_index(level=index_name, drop=False)
        else:
            res._supplementary_data = pd.DataFrame(index=self._mass_data.index, columns=[index_name],
                                                   data=self._mass_data.index.get_level_values(index_name))

        return res

    def _get_component_ranges(self, ranges: dict[str, list]) -> dict[str, list]:

        d_ranges: dict = get_column_config(config_dict=self.config, var_map=self.variable_map,
                                           config_key='range')
        # filter to include only components
        d_ranges = {k: v for k, v in d_ranges.items() if k in self.composition_columns}

        # modify the default dict based on any user passed constraints
        if ranges:
            for k, v in ranges.items():
                d_ranges[k] = v

        return d_ranges


class OutOfRangeStatus:
    """A class to check and report out-of-range records in an MC object."""

    def __init__(self, mc: 'MC', ranges: dict[str, list]):
        """Initialize with an MC object."""
        self._logger = logging.getLogger(__name__)
        self.mc: 'MC' = mc
        self.ranges: Optional[dict[str, list]] = None
        self.oor: Optional[pd.DataFrame] = None
        self.num_oor: Optional[int] = None
        self.failing_components: Optional[list[str]] = None

        if mc.mass_data is not None:
            self.ranges = self.get_ranges(ranges)
            self.oor: pd.DataFrame = self._check_range()
            self.num_oor: int = len(self.oor)
            self.failing_components: Optional[list[str]] = list(
                self.oor.dropna(axis=1, how='all').columns) if self.num_oor > 0 else None

    def get_ranges(self, ranges: dict[str, list]) -> dict[str, list]:

        d_ranges: dict = get_column_config(config_dict=self.mc.config, var_map=self.mc.variable_map,
                                           config_key='range')

        # modify the default dict based on any user passed constraints
        if ranges:
            for k, v in ranges.items():
                d_ranges[k] = v

        return d_ranges

    def _check_range(self) -> pd.DataFrame:
        """Check if all records are within the constraints."""
        if self.mc._mass_data is not None:
            df: pd.DataFrame = self.mc.data[self.ranges.keys()]
            chunks = []
            for variable, bounds in self.ranges.items():
                chunks.append(df.loc[(df[variable] < bounds[0]) | (df[variable] > bounds[1]), variable])
            oor: pd.DataFrame = pd.concat(chunks, axis='columns')
        else:  # An empty object will have ok status
            oor: pd.DataFrame = pd.DataFrame(columns=list(self.ranges.keys()))
        return oor

    @property
    def ok(self) -> bool:
        """Return True if all records are within range, False otherwise."""
        if self.num_oor > 0:
            self._logger.warning(f'{self.num_oor} out of range records exist.')
        return True if self.num_oor == 0 else False

    def __str__(self) -> str:
        """Return a string representation of the status."""
        res: str = f'status.ok: {self.ok}\n'
        res += f'num_oor: {self.num_oor}'
        return res

    def __eq__(self, other: object) -> bool:
        """Return True if other Status has the same out-of-range records."""
        if isinstance(other, OutOfRangeStatus):
            return self.oor.equals(other.oor)
        return False


class SampleStatus:
    """A class to check and report sample status in an MC object.

    A MassComposition object (Sample, Stream, etc) can be unhealthy if:
    1. the total mass of components is greater than the dry mass
    2. any of the masses are negative
    3. the component values are out of range
    """

    def __init__(self, mc: 'MC', component_limits: dict):
        """Initialize with an MC object."""
        self._logger = logging.getLogger(__name__)
        self.mc: 'MC' = mc
        self.sample_status: Optional[dict] = None
        self.num_samples: Optional[int] = None
        self.failing_samples: Optional[list[str]] = None

        if mc.mass_data is not None:
            self.sample_status = self.get_sample_status(component_limits)
            self.num_samples = len(self.sample_status)
            self.failing_samples = [sample for sample, status in self.sample_status.items() if not status['ok']]

    def get_sample_status(self, component_limits: dict) -> dict:
        """Check if all records are within the constraints."""
        sample_status = {}
        if self.mc._mass_data is not None:
            df: pd.DataFrame = self.mc.data[self.mc.composition_columns]
            mass_dry = self.mc._mass_data[self.mc.mass_dry_var]

            # Check for component limits
            for component, limits in component_limits.items():
                oor = df[(df[component] < limits[0]) | (df[component] > limits[1])]
                for sample in oor.index:
                    if sample not in sample_status:
                        sample_status[sample] = {'ok': True, 'causes': []}
                    sample_status[sample]['ok'] = False
                    sample_status[sample]['causes'].append(f"{component} out of range")

            # Check if total mass of components is greater than dry mass
            total_component_mass = df.sum(axis=1)
            for sample in total_component_mass.index:
                if sample not in sample_status:
                    sample_status[sample] = {'ok': True, 'causes': []}
                if total_component_mass[sample] > mass_dry[sample]:
                    sample_status[sample]['ok'] = False
                    sample_status[sample]['causes'].append("Total mass of components greater than dry mass")

            # Check for negative masses
            negative_masses = df[df < 0].dropna(how='all')
            for sample in negative_masses.index:
                if sample not in sample_status:
                    sample_status[sample] = {'ok': True, 'causes': []}
                sample_status[sample]['ok'] = False
                sample_status[sample]['causes'].append("Negative mass detected")
        return sample_status

    @property
    def ok(self) -> bool:
        """Return True if all records are within range, False otherwise."""
        if self.num_samples > 0:
            self._logger.warning(f'{self.num_samples} unhealthy samples exist.')
        return True if self.num_samples == 0 else False

    def __str__(self) -> str:
        """Return a string representation of the status."""
        res: str = f'status.ok: {self.ok}\n'
        res += f'num_samples: {self.num_samples}'
        return res

    def __eq__(self, other: object) -> bool:
        """Return True if other Status has the same out-of-range records."""
        if isinstance(other, SampleStatus):
            return self.sample_status == other.sample_status
        return False



import copy
import json
import logging
import uuid
from pathlib import Path
from typing import Dict, List, Optional, Tuple, Union, TypeVar, TYPE_CHECKING
import re

import matplotlib
import matplotlib.cm as cm
import networkx as nx
import numpy as np
import pandas as pd
import plotly.graph_objects as go
import seaborn as sns
import yaml
from matplotlib import pyplot as plt
from matplotlib.colors import ListedColormap, LinearSegmentedColormap
from networkx.algorithms.dag import is_directed_acyclic_graph
from plotly.subplots import make_subplots

from elphick.geomet import Sample
from elphick.geomet.base import MC
from elphick.geomet.config.config_read import get_column_config
from elphick.geomet.flowsheet.operation import NodeType, OP, PartitionOperation, Operation
from elphick.geomet.plot import parallel_plot, comparison_plot
from elphick.geomet.utils.layout import digraph_linear_layout
from elphick.geomet.flowsheet.loader import streams_from_dataframe

# if TYPE_CHECKING:
from elphick.geomet.flowsheet.stream import Stream

# generic type variable, used for type hinting that play nicely with subclasses
FS = TypeVar('FS', bound='Flowsheet')


class Flowsheet:
    def __init__(self, name: str = 'Flowsheet'):
        self.graph: nx.DiGraph = nx.DiGraph(name=name)
        self._logger: logging.Logger = logging.getLogger(__class__.__name__)

    @property
    def name(self) -> str:
        return self.graph.name

    @name.setter
    def name(self, value: str):
        self.graph.name = value

    @property
    def healthy(self) -> bool:
        return self.all_nodes_healthy and self.all_streams_healthy

    @property
    def all_nodes_healthy(self) -> bool:
        bal_vals: List = [self.graph.nodes[n]['mc'].is_balanced for n in self.graph.nodes]
        bal_vals = [bv for bv in bal_vals if bv is not None]
        return all(bal_vals)

    @property
    def all_streams_healthy(self) -> bool:
        """Check if all streams are healthy"""
        # account for the fact that some edges may not have an mc object
        if not all([d['mc'] for u, v, d in self.graph.edges(data=True)]):
            return False
        return all([self.graph.edges[u, v]['mc'].status.ok for u, v in self.graph.edges])

    @classmethod
    def from_objects(cls, objects: list[MC],
                     name: Optional[str] = 'Flowsheet') -> FS:
        """Instantiate from a list of objects

        This method is only suitable for objects that have the `_nodes` property set, such as objects that have
        been created from math operations, which preserve relationships between objects (via the nodes property)

        Args:
            objects: List of MassComposition objects, such as Sample, IntervalSample, Stream or BlockModel
            name: name of the flowsheet/network

        Returns:

        """
        from elphick.geomet.flowsheet.operation import Operation
        cls._check_indexes(objects)
        bunch_of_edges: list = []
        for stream in objects:
            if stream.nodes is None:
                raise KeyError(f'Stream {stream.name} does not have the node property set')
            nodes = stream.nodes

            # add the objects to the edges
            bunch_of_edges.append((nodes[0], nodes[1], {'mc': stream, 'name': stream.name}))

        graph = nx.DiGraph(name=name)
        graph.add_edges_from(bunch_of_edges)
        operation_objects: dict = {}
        for node in graph.nodes:
            operation_objects[node] = Operation(name=node)
        nx.set_node_attributes(graph, operation_objects, 'mc')

        for node in graph.nodes:
            operation_objects[node].inputs = [graph.get_edge_data(e[0], e[1])['mc'] for e in graph.in_edges(node)]
            operation_objects[node].outputs = [graph.get_edge_data(e[0], e[1])['mc'] for e in graph.out_edges(node)]

        graph = nx.convert_node_labels_to_integers(graph)
        # update the temporary nodes on the mc object property to match the renumbered integers
        for node1, node2, data in graph.edges(data=True):
            data['mc'].nodes = [node1, node2]
        # update the node names after renumbering
        for node in graph.nodes:
            graph.nodes[node]['mc'].name = str(node)
        obj = cls()
        obj.graph = graph
        return obj

    @classmethod
    def from_dataframe(cls, df: pd.DataFrame, name: Optional[str] = 'Flowsheet',
                       mc_name_col: Optional[str] = None, n_jobs: int = 1) -> FS:
        """Instantiate from a DataFrame

        Args:
            df: The DataFrame
            name: name of the network
            mc_name_col: The column specified contains the names of objects to create.
              If None the DataFrame is assumed to be wide and the mc objects will be extracted from column prefixes.
            n_jobs: The number of parallel jobs to run.  If -1, will use all available cores.

        Returns:
            Flowsheet: An instance of the Flowsheet class initialized from the provided DataFrame.

        """
        streams: list[Sample] = streams_from_dataframe(df=df, mc_name_col=mc_name_col, n_jobs=n_jobs)
        return cls().from_objects(objects=streams, name=name)

    @classmethod
    def from_dict(cls, config: dict) -> FS:
        """Create a flowsheet from a dictionary

        Args:
            config: dictionary containing the flowsheet configuration

        Returns:
            A Flowsheet object with no data on the edges
        """

        from elphick.geomet.flowsheet.operation import Operation

        if 'FLOWSHEET' not in config:
            raise ValueError("Dictionary does not contain 'FLOWSHEET' root node")

        flowsheet_config = config['FLOWSHEET']

        # create the Stream objects
        bunch_of_edges: list = []
        for stream, stream_config in flowsheet_config['streams'].items():
            bunch_of_edges.append(
                (stream_config['node_in'], stream_config['node_out'], {'mc': None, 'name': stream_config['name']}))

        graph = nx.DiGraph(name=flowsheet_config['flowsheet']['name'])
        graph.add_edges_from(bunch_of_edges)
        operation_objects: dict = {}
        for node in graph.nodes:
            # create the correct type of node object
            if node in flowsheet_config['operations']:
                operation_type = flowsheet_config['operations'][node].get('type', 'Operation')
                if operation_type == 'PartitionOperation':
                    # get the output stream names from the graph
                    output_stream_names = [d['name'] for u, v, d in graph.out_edges(node, data=True)]
                    node_config = flowsheet_config['operations'][node]
                    node_config['output_stream_names'] = output_stream_names
                    operation_objects[node] = PartitionOperation.from_dict(node_config)
                else:
                    operation_objects[node] = Operation.from_dict(flowsheet_config['operations'][node])

                # set the input and output streams on the operation object for the selected node
                operation_objects[node].inputs = [graph.get_edge_data(e[0], e[1])['mc'] for e in graph.in_edges(node)]
                operation_objects[node].outputs = [graph.get_edge_data(e[0], e[1])['mc'] for e in graph.out_edges(node)]

        nx.set_node_attributes(graph, operation_objects, 'mc')

        graph = nx.convert_node_labels_to_integers(graph)

        obj = cls()
        obj.graph = graph

        return obj

    # @classmethod
    # def from_dict_todo(cls, config: dict) -> FS:
    # TODO: This method is not yet implemented - fails because the Operations do not have inputs or outputs set.
    #     flowsheet = cls()
    #
    #     # Process streams
    #     for stream_name, stream_data in config['FLOWSHEET']['streams'].items():
    #         stream = Stream.from_dict(stream_data)
    #         flowsheet.add_stream(stream)
    #
    #     # Process operations
    #     for operation_name, operation_data in config['FLOWSHEET']['operations'].items():
    #         operation_type = operation_data.get('type', 'Operation')
    #         if operation_type == 'PartitionOperation':
    #             operation = PartitionOperation.from_dict(operation_data)
    #         else:
    #             operation = Operation.from_dict(operation_data)
    #         flowsheet.add_operation(operation)
    #
    #     return flowsheet

    @classmethod
    def from_yaml(cls, file_path: Path) -> FS:
        """Create a flowsheet from yaml

        Args:
            file_path: path to the yaml file

        Returns:
            A Flowsheet object with no data on the edges
        """
        with open(file_path, 'r') as file:
            config = yaml.safe_load(file)

        return cls.from_dict(config)

    @classmethod
    def from_json(cls, file_path: Path) -> FS:
        """Create a flowsheet from json

        Args:
            file_path: path to the json file

        Returns:
            A Flowsheet object with no data on the edges
        """
        with open(file_path, 'r') as file:
            config = json.load(file)

        return cls.from_dict(config)

    def add_stream(self, stream: 'Stream'):
        """Add a stream to the flowsheet."""
        self.graph.add_edge(stream.nodes[0], stream.nodes[1], mc=stream, name=stream.name)

    def add_operation(self, operation: 'Operation'):
        """Add an operation to the flowsheet."""
        self.graph.add_node(operation.name, mc=operation)

    def unhealthy_stream_records(self) -> pd.DataFrame:
        """Return on unhealthy streams

        Return the records for all streams that are not healthy.
        Returns:
            DataFrame: A DataFrame containing the unhealthy stream records
        """
        unhealthy_edges = [e for e in self.graph.edges if not self.graph.edges[e]['mc'].status.ok]
        unhealthy_data: pd.DataFrame = pd.concat(
            [self.graph.edges[e]['mc'].status.oor.assign(stream=self.graph.edges[e]['mc'].name) for e in
             unhealthy_edges], axis=1)
        # move the last column to the front
        unhealthy_data = unhealthy_data[[unhealthy_data.columns[-1]] + list(unhealthy_data.columns[:-1])]

        # append the flowsheet records for additional context
        records: pd.DataFrame = self.to_dataframe()
        records = records.unstack(level='name').swaplevel(axis=1).sort_index(axis=1, level=0, sort_remaining=False)
        records.columns = [f"{col[0]}_{col[1]}" for col in records.columns]

        result = unhealthy_data.merge(records, left_index=True, right_index=True, how='left')
        return result

    def unhealthy_node_records(self) -> pd.DataFrame:
        """Return unhealthy nodes

        Return the records for all nodes that are not healthy.
        Returns:
            DataFrame: A DataFrame containing the unhealthy node records
        """
        unhealthy_nodes = [n for n in self.graph.nodes if
                           self.graph.nodes[n]['mc'].node_type == NodeType.BALANCE and not self.graph.nodes[n][
                               'mc'].is_balanced]
        unhealthy_data: pd.DataFrame = pd.concat(
            [self.graph.nodes[n]['mc'].unbalanced_records.assign(node=self.graph.nodes[n]['mc'].name) for n in
             unhealthy_nodes], axis=1)
        # move the last column to the front
        unhealthy_data = unhealthy_data[[unhealthy_data.columns[-1]] + list(unhealthy_data.columns[:-1])]

        # todo: append  the streams around the node

        return unhealthy_data

    def copy_without_stream_data(self):
        """Copy without stream data"""
        new_flowsheet = Flowsheet(name=self.name)
        new_graph = nx.DiGraph()

        # Copy nodes with Operation objects
        for node, data in self.graph.nodes(data=True):
            new_data = data.copy()
            new_graph.add_node(node, **new_data)

        # Copy edges with mc attribute set to None
        for u, v, data in self.graph.edges(data=True):
            new_data = {k: (None if k == 'mc' else copy.deepcopy(v)) for k, v in data.items()}
            new_graph.add_edge(u, v, **new_data)

        new_flowsheet.graph = new_graph
        return new_flowsheet

    def solve(self):
        """Solve missing streams"""

        if not is_directed_acyclic_graph(self.graph):
            self._logger.error("Graph is not a Directed Acyclic Graph (DAG), so cannot be solved.")
            self._logger.debug(f"Graph nodes: {self.graph.nodes(data=True)}")
            self._logger.debug(f"Graph edges: {self.graph.edges(data=True)}")
            raise ValueError("Graph is not a Directed Acyclic Graph (DAG), so cannot be solved.")

        # Check the number of missing mc's on edges in the network
        missing_count: int = sum([1 for u, v, d in self.graph.edges(data=True) if d['mc'] is None])
        prev_missing_count = missing_count + 1  # Initialize with a value greater than missing_count

        while 0 < missing_count < prev_missing_count:
            prev_missing_count = missing_count
            for node in nx.topological_sort(self.graph):
                if self.graph.nodes[node]['mc'].node_type == NodeType.BALANCE:
                    if self.graph.nodes[node]['mc'].has_empty_input:
                        mc: MC = self.graph.nodes[node]['mc'].solve()
                        # copy the solved object to the empty input edges
                        for predecessor in self.graph.predecessors(node):
                            edge_data = self.graph.get_edge_data(predecessor, node)
                            if edge_data and edge_data['mc'] is None:
                                edge_data['mc'] = mc
                                edge_data['mc'].name = edge_data['name']
                                self.set_operation_data(predecessor)

                    if self.graph.nodes[node]['mc'].has_empty_output:
                        # There are two cases to be managed, 1. a single output missing,
                        # 2. a partition operation that returns two outputs
                        if isinstance(self.graph.nodes[node]['mc'], PartitionOperation):
                            partition_stream: str = self.graph.nodes[node]['mc'].partition['partition_stream']
                            mc1, mc2 = self.graph.nodes[node]['mc'].solve()
                            # copy the solved object to the empty output edges
                            for successor in self.graph.successors(node):
                                edge_data = self.graph.get_edge_data(node, successor)
                                if edge_data and edge_data['mc'] is None:
                                    edge_data['mc'] = mc1 if edge_data['name'] == partition_stream else mc2
                                    edge_data['mc'].name = edge_data['name']
                                    self.set_operation_data(successor)

                        else:
                            mc: MC = self.graph.nodes[node]['mc'].solve()
                            # copy the solved object to the empty output edges
                            for successor in self.graph.successors(node):
                                edge_data = self.graph.get_edge_data(node, successor)
                                if edge_data and edge_data['mc'] is None:
                                    edge_data['mc'] = mc
                                    edge_data['mc'].name = edge_data['name']
                                    self.set_operation_data(successor)
                        self.set_operation_data(node)

            missing_count: int = sum([1 for u, v, d in self.graph.edges(data=True) if d['mc'] is None])
            self._logger.info(f"Missing count: {missing_count}")

        if missing_count > 0:
            self._logger.error(f"Failed to solve the flowsheet. Missing count: {missing_count}")
            raise ValueError(
                f"Failed to solve the flowsheet. Some streams are still missing. Missing count: {missing_count}")

    def query(self, expr: str, stream_name: Optional[str] = None, inplace=False) -> 'Flowsheet':
        """Reduce the Flowsheet Stream records with a query

        Args:
            expr: The query string to apply to all streams. The query is applied in place. The LHS of the
                expression requires a prefix that defines the stream name e.g. stream_name.var_name > 0.5
            stream_name: The name of the stream to apply the query to. If None, the query is applied to the
                first input stream.
            inplace: If True, apply the query in place on the same object, otherwise return a new instance.

        Returns:
            A Flowsheet object where the stream records conform to the query
        """
        if stream_name is None:
            input_stream: MC = self.get_input_streams()[0]
        else:
            input_stream: MC = self.get_stream_by_name(name=stream_name)
        filtered_index: pd.Index = input_stream.data.query(expr).index
        return self._filter(filtered_index, inplace)

    def filter_by_index(self, index: pd.Index, inplace: bool = False) -> 'Flowsheet':
        """Filter the Flowsheet Stream records by a given index.

        Args:
            index: The index to filter the data.
            inplace: If True, apply the filter in place on the same object, otherwise return a new instance.

        Returns:
            A Flowsheet object where the stream records are filtered by the given index.
        """
        return self._filter(index, inplace)

    def _filter(self, index: pd.Index, inplace: bool = False) -> 'Flowsheet':
        """Private method to filter the Flowsheet Stream records by a given index.

        Args:
            index: The index to filter the data.
            inplace: If True, apply the filter in place on the same object, otherwise return a new instance.

        Returns:
            A Flowsheet object where the stream records are filtered by the given index.
        """
        if inplace:
            for u, v, d in self.graph.edges(data=True):
                if d.get('mc') is not None:
                    d.get('mc').filter_by_index(index)
            return self
        else:
            obj: Flowsheet = self.copy_without_stream_data()
            for u, v, d in self.graph.edges(data=True):
                if d.get('mc') is not None:
                    mc: MC = d.get('mc')
                    mc_new = mc.__class__(name=mc.name)
                    # Copy each attribute
                    for attr, value in mc.__dict__.items():
                        if attr in ['_mass_data', '_supplementary_data'] and value is not None:
                            value = value.loc[index]
                        setattr(mc_new, attr, copy.deepcopy(value))
                    mc_new.aggregate = mc_new.weight_average()
                    obj.graph[u][v]['mc'] = mc_new
            return obj

    def get_input_streams(self) -> list[MC]:
        """Get the input (feed) streams (edge objects)

        Returns:
            List of MassComposition-like objects
        """

        # Create a dictionary that maps node names to their degrees
        degrees = {n: d for n, d in self.graph.degree()}

        res: list[MC] = [d['mc'] for u, v, d in self.graph.edges(data=True) if degrees[u] == 1]
        if not res:
            raise ValueError("No input streams found")
        return res

    def get_output_streams(self) -> list[MC]:
        """Get the output (product) streams (edge objects)

        Returns:
            List of MassComposition-like objects
        """

        # Create a dictionary that maps node names to their degrees
        degrees = {n: d for n, d in self.graph.degree()}

        res: list[MC] = [d['mc'] for u, v, d in self.graph.edges(data=True) if degrees[v] == 1]
        if not res:
            raise ValueError("No output streams found")
        return res

    @staticmethod
    def _check_indexes(streams):

        list_of_indexes = [s._mass_data.index for s in streams]
        types_of_indexes = [type(i) for i in list_of_indexes]
        # check the index types are consistent
        if len(set(types_of_indexes)) != 1:
            raise KeyError("stream index types are not consistent")

    def plot(self, orientation: str = 'horizontal') -> plt.Figure:
        """Plot the network with matplotlib

        Args:
            orientation: 'horizontal'|'vertical' network layout

        Returns:

        """

        hf, ax = plt.subplots()
        # pos = nx.spring_layout(self, seed=1234)
        pos = digraph_linear_layout(self.graph, orientation=orientation)

        edge_labels: Dict = {}
        edge_colors: List = []
        node_colors: List = []

        for node1, node2, data in self.graph.edges(data=True):
            edge_labels[(node1, node2)] = data['mc'].name if data['mc'] is not None else data['name']
            if data['mc'] and data['mc'].status.ok:
                edge_colors.append('gray')
            else:
                edge_colors.append('red')

        for n in self.graph.nodes:
            if self.graph.nodes[n]['mc'].node_type == NodeType.BALANCE:
                if self.graph.nodes[n]['mc'].is_balanced:
                    node_colors.append('green')
                else:
                    node_colors.append('red')
            else:
                node_colors.append('gray')

        nx.draw(self.graph, pos=pos, ax=ax, with_labels=True, font_weight='bold',
                node_color=node_colors, edge_color=edge_colors)

        nx.draw_networkx_edge_labels(self.graph, pos=pos, ax=ax, edge_labels=edge_labels, font_color='black')
        ax.set_title(self._plot_title(html=False), fontsize=10)

        return hf

    def _plot_title(self, html: bool = True, compact: bool = False):
        # title = self.name
        title = (f"{self.name}<br><sup>Nodes Healthy: "
                 f"<span style='color: {'red' if not self.all_nodes_healthy else 'black'}'>{self.all_nodes_healthy}</span>, "
                 f"Streams Healthy: "
                 f"<span style='color: {'red' if not self.all_streams_healthy else 'black'}'>{self.all_streams_healthy}</span></sup>")
        # if compact:
        #     title = title.replace("<br><br>", "<br>").replace("<br>Edge", ", Edge")
        # if not self.edge_status[0]:
        #     title = title.replace("</sup>", "") + f", {self.edge_status[1]}</sup>"
        if not html:
            title = title.replace('<br><br>', '\n').replace('<br>', '\n').replace('<sup>', '').replace('</sup>', '')
            title = re.sub(r'<span style=.*?>(.*?)</span>', r'\1', title)
        return title

    def report(self, apply_formats: bool = False) -> pd.DataFrame:
        """Summary Report

        Total Mass and weight averaged composition
        Returns:

        """
        chunks: List[pd.DataFrame] = []
        for n, nbrs in self.graph.adj.items():
            for nbr, eattr in nbrs.items():
                if eattr['mc'] is None or eattr['mc'].data.empty:
                    edge_name: str = eattr['mc']['name'] if eattr['mc'] is not None else eattr['name']
                    raise KeyError(f"Cannot generate report on empty dataset: {edge_name}")
                chunks.append(eattr['mc'].aggregate.assign(name=eattr['mc'].name))
        rpt: pd.DataFrame = pd.concat(chunks, axis='index').set_index('name')
        if apply_formats:
            fmts: Dict = self._get_column_formats(rpt.columns)
            for k, v in fmts.items():
                rpt[k] = rpt[k].apply((v.replace('%', '{:,') + '}').format)
        return rpt

    def _get_column_formats(self, columns: List[str], strip_percent: bool = False) -> Dict[str, str]:
        """

        Args:
            columns: The columns to lookup format strings for
            strip_percent: If True remove the leading % symbol from the format (for plotly tables)

        Returns:

        """
        strm = self.get_input_streams()[0]
        d_format: dict = get_column_config(config_dict=strm.config, var_map=strm.variable_map, config_key='format')

        if strip_percent:
            d_format = {k: v.strip('%') for k, v in d_format.items()}

        return d_format

    def plot_balance(self, facet_col_wrap: int = 3,
                     color: Optional[str] = 'node') -> go.Figure:
        """Plot input versus output across all nodes in the network

        Args:
            facet_col_wrap: the number of subplots per row before wrapping
            color: The optional variable to color by. If None color will be by Node

        Returns:

        """
        # prepare the data
        chunks_in: List = []
        chunks_out: List = []
        for n in self.graph.nodes:
            if self.graph.nodes[n]['mc'].node_type == NodeType.BALANCE:
                chunks_in.append(self.graph.nodes[n]['mc'].add('in').assign(**{'direction': 'in', 'node': n}))
                chunks_out.append(self.graph.nodes[n]['mc'].add('out').assign(**{'direction': 'out', 'node': n}))
        df_in: pd.DataFrame = pd.concat(chunks_in)
        index_names = ['direction', 'node'] + df_in.index.names
        df_in = df_in.reset_index().melt(id_vars=index_names)
        df_out: pd.DataFrame = pd.concat(chunks_out).reset_index().melt(id_vars=index_names)
        df_plot: pd.DataFrame = pd.concat([df_in, df_out])
        df_plot = df_plot.set_index(index_names + ['variable'], append=True).unstack(['direction'])
        df_plot.columns = df_plot.columns.droplevel(0)
        df_plot.reset_index(level=list(np.arange(-1, -len(index_names) - 1, -1)), inplace=True)
        df_plot['node'] = pd.Categorical(df_plot['node'])

        # plot
        fig = comparison_plot(data=df_plot,
                              x='in', y='out',
                              facet_col_wrap=facet_col_wrap,
                              color=color)
        return fig

    def plot_network(self, orientation: str = 'horizontal') -> go.Figure:
        """Plot the network with plotly

        Args:
            orientation: 'horizontal'|'vertical' network layout

        Returns:

        """
        # pos = nx.spring_layout(self, seed=1234)
        pos = digraph_linear_layout(self.graph, orientation=orientation)

        edge_traces, node_trace, edge_annotation_trace = self._get_scatter_node_edges(pos)
        title = self._plot_title()

        fig = go.Figure(data=[*edge_traces, node_trace, edge_annotation_trace],
                        layout=go.Layout(
                            title=title,
                            titlefont_size=16,
                            showlegend=False,
                            hovermode='closest',
                            margin=dict(b=20, l=5, r=5, t=40),
                            xaxis=dict(showgrid=False, zeroline=False, showticklabels=False),
                            yaxis=dict(showgrid=False, zeroline=False, showticklabels=False),
                            paper_bgcolor='rgba(0,0,0,0)',
                            plot_bgcolor='rgba(0,0,0,0)'
                        ),
                        )
        # for k, d_args in edge_annotations.items():
        #     fig.add_annotation(x=d_args['pos'][0], y=d_args['pos'][1], text=k, textangle=d_args['angle'])

        return fig

    def plot_sankey(self,
                    width_var: str = 'mass_dry',
                    color_var: Optional[str] = None,
                    edge_colormap: Optional[str] = 'copper_r',
                    vmin: Optional[float] = None,
                    vmax: Optional[float] = None,
                    ) -> go.Figure:
        """Plot the Network as a sankey

        Args:
            width_var: The variable that determines the sankey width
            color_var: The optional variable that determines the sankey edge color
            edge_colormap: The optional colormap.  Used with color_var.
            vmin: The value that maps to the minimum color
            vmax: The value that maps to the maximum color

        Returns:

        """
        # Create a mapping of node names to indices, and the integer nodes
        node_indices = {node: index for index, node in enumerate(self.graph.nodes)}
        int_graph = nx.relabel_nodes(self.graph, node_indices)

        # Generate the sankey diagram arguments using the new graph with integer nodes
        d_sankey = self._generate_sankey_args(int_graph, color_var, edge_colormap, width_var, vmin, vmax)

        # Create the sankey diagram
        node, link = self._get_sankey_node_link_dicts(d_sankey)
        fig = go.Figure(data=[go.Sankey(node=node, link=link)])
        title = self._plot_title()
        fig.update_layout(title_text=title, font_size=10)
        return fig

    def table_plot(self,
                   plot_type: str = 'sankey',
                   cols_exclude: Optional[List] = None,
                   table_pos: str = 'left',
                   table_area: float = 0.4,
                   table_header_color: str = 'cornflowerblue',
                   table_odd_color: str = 'whitesmoke',
                   table_even_color: str = 'lightgray',
                   sankey_width_var: str = 'mass_dry',
                   sankey_color_var: Optional[str] = None,
                   sankey_edge_colormap: Optional[str] = 'copper_r',
                   sankey_vmin: Optional[float] = None,
                   sankey_vmax: Optional[float] = None,
                   network_orientation: Optional[str] = 'horizontal'
                   ) -> go.Figure:
        """Plot with table of edge averages

        Args:
            plot_type: The type of plot ['sankey', 'network']
            cols_exclude: List of columns to exclude from the table
            table_pos: Position of the table ['left', 'right', 'top', 'bottom']
            table_area: The proportion of width or height to allocate to the table [0, 1]
            table_header_color: Color of the table header
            table_odd_color: Color of the odd table rows
            table_even_color: Color of the even table rows
            sankey_width_var: If plot_type is sankey, the variable that determines the sankey width
            sankey_color_var: If plot_type is sankey, the optional variable that determines the sankey edge color
            sankey_edge_colormap: If plot_type is sankey, the optional colormap.  Used with sankey_color_var.
            sankey_vmin: The value that maps to the minimum color
            sankey_vmax: The value that maps to the maximum color
            network_orientation: The orientation of the network layout 'vertical'|'horizontal'

        Returns:

        """

        valid_plot_types: List[str] = ['sankey', 'network']
        if plot_type not in valid_plot_types:
            raise ValueError(f'The supplied plot_type is not in {valid_plot_types}')

        valid_table_pos: List[str] = ['top', 'bottom', 'left', 'right']
        if table_pos not in valid_table_pos:
            raise ValueError(f'The supplied table_pos is not in {valid_table_pos}')

        d_subplot, d_table, d_plot = self._get_position_kwargs(table_pos, table_area, plot_type)

        fig = make_subplots(**d_subplot, print_grid=False)

        df: pd.DataFrame = self.report().reset_index()
        if cols_exclude:
            df = df[[col for col in df.columns if col not in cols_exclude]]
        fmt: List[str] = ['%s'] + list(self._get_column_formats(df.columns, strip_percent=True).values())
        column_widths = [2] + [1] * (len(df.columns) - 1)

        fig.add_table(
            header=dict(values=list(df.columns),
                        fill_color=table_header_color,
                        align='center',
                        font=dict(color='black', size=12)),
            columnwidth=column_widths,
            cells=dict(values=df.transpose().values.tolist(),
                       align='left', format=fmt,
                       fill_color=[
                           [table_odd_color if i % 2 == 0 else table_even_color for i in range(len(df))] * len(
                               df.columns)]),
            **d_table)

        if plot_type == 'sankey':
            # Create a mapping of node names to indices, and the integer nodes
            node_indices = {node: index for index, node in enumerate(self.graph.nodes)}
            int_graph = nx.relabel_nodes(self.graph, node_indices)

            # Generate the sankey diagram arguments using the new graph with integer nodes
            d_sankey = self._generate_sankey_args(int_graph, sankey_color_var,
                                                  sankey_edge_colormap,
                                                  sankey_width_var,
                                                  sankey_vmin,
                                                  sankey_vmax)
            node, link = self._get_sankey_node_link_dicts(d_sankey)
            fig.add_trace(go.Sankey(node=node, link=link), **d_plot)

        elif plot_type == 'network':
            # pos = nx.spring_layout(self, seed=1234)
            pos = digraph_linear_layout(self.graph, orientation=network_orientation)

            edge_traces, node_trace, edge_annotation_trace = self._get_scatter_node_edges(pos)
            fig.add_traces(data=[*edge_traces, node_trace, edge_annotation_trace], **d_plot)

            fig.update_layout(showlegend=False, hovermode='closest',
                              xaxis=dict(showgrid=False, zeroline=False, showticklabels=False),
                              yaxis=dict(showgrid=False, zeroline=False, showticklabels=False),
                              paper_bgcolor='rgba(0,0,0,0)',
                              plot_bgcolor='rgba(0,0,0,0)'
                              )

        title = self._plot_title(compact=True)
        fig.update_layout(title_text=title, font_size=12)

        return fig

    def to_dataframe(self, stream_names: Optional[list[str]] = None, tidy: bool = True,
                     as_mass: bool = False) -> pd.DataFrame:
        """Return a tidy dataframe

        Adds the mc name to the index so indexes are unique.

        Args:
            stream_names: Optional List of names of Stream/MassComposition objects (network edges) for export
            tidy: If True, the data will be returned in a tidy format, otherwise wide
            as_mass: If True, the mass data will be returned instead of the mass-composition data

        Returns:

        """
        chunks: List[pd.DataFrame] = []
        for u, v, data in self.graph.edges(data=True):
            if (stream_names is None) or ((stream_names is not None) and (data['mc'].name in stream_names)):
                if as_mass:
                    chunks.append(data['mc'].mass_data.assign(name=data['mc'].name))
                else:
                    chunks.append(data['mc'].data.assign(name=data['mc'].name))

        results: pd.DataFrame = pd.concat(chunks, axis='index').set_index('name', append=True)
        if not tidy:  # wide format
            results = results.unstack(level='name')
            column_order: list[str] = [f'{name}_{attr}' for name in results.columns.levels[1] for attr in
                                       results.columns.levels[0]]
            results.columns = [f'{col[1]}_{col[0]}' for col in results.columns]
            results = results[column_order]

        return results

    def plot_parallel(self,
                      names: Optional[str] = None,
                      color: Optional[str] = None,
                      vars_include: Optional[List[str]] = None,
                      vars_exclude: Optional[List[str]] = None,
                      title: Optional[str] = None,
                      include_dims: Optional[Union[bool, List[str]]] = True,
                      plot_interval_edges: bool = False) -> go.Figure:
        """Create an interactive parallel plot

        Useful to explore multidimensional data like mass-composition data

        Args:
            names: Optional List of Names to plot
            color: Optional color variable
            vars_include: Optional List of variables to include in the plot
            vars_exclude: Optional List of variables to exclude in the plot
            title: Optional plot title
            include_dims: Optional boolean or list of dimension to include in the plot.  True will show all dims.
            plot_interval_edges: If True, interval edges will be plotted instead of interval mid

        Returns:

        """
        df: pd.DataFrame = self.to_dataframe(stream_names=names)

        if not title and hasattr(self, 'name'):
            title = self.name

        fig = parallel_plot(data=df, color=color, vars_include=vars_include, vars_exclude=vars_exclude, title=title,
                            include_dims=include_dims, plot_interval_edges=plot_interval_edges)
        return fig

    def _generate_sankey_args(self, int_graph, color_var, edge_colormap, width_var, v_min, v_max):
        rpt: pd.DataFrame = self.report()
        if color_var is not None:
            cmap = sns.color_palette(edge_colormap, as_cmap=True)
            rpt: pd.DataFrame = self.report()
            if not v_min:
                v_min = np.floor(rpt[color_var].min())
            if not v_max:
                v_max = np.ceil(rpt[color_var].max())

        # run the report for the hover data
        d_custom_data: Dict = self._rpt_to_html(df=rpt)
        source: List = []
        target: List = []
        value: List = []
        edge_custom_data = []
        edge_color: List = []
        edge_labels: List = []
        node_colors: List = []
        node_labels: List = []

        for n in int_graph.nodes:
            node_labels.append(int_graph.nodes[n]['mc'].name)

            if int_graph.nodes[n]['mc'].node_type == NodeType.BALANCE:
                if int_graph.nodes[n]['mc'].is_balanced:
                    node_colors.append('green')
                else:
                    node_colors.append('red')
            else:
                node_colors.append('blue')

        for u, v, data in int_graph.edges(data=True):
            edge_labels.append(data['mc'].name)
            source.append(u)
            target.append(v)
            value.append(float(data['mc'].aggregate[width_var].iloc[0]))
            edge_custom_data.append(d_custom_data[data['mc'].name])

            if color_var is not None:
                val: float = float(data['mc'].aggregate[color_var].iloc[0])
                str_color: str = f'rgba{self._color_from_float(v_min, v_max, val, cmap)}'
                edge_color.append(str_color)
            else:
                edge_color: Optional[str] = None

        d_sankey: Dict = {'node_color': node_colors,
                          'edge_color': edge_color,
                          'edge_custom_data': edge_custom_data,
                          'edge_labels': edge_labels,
                          'labels': node_labels,
                          'source': source,
                          'target': target,
                          'value': value}

        return d_sankey

    @staticmethod
    def _get_sankey_node_link_dicts(d_sankey: Dict):
        node: Dict = dict(
            pad=15,
            thickness=20,
            line=dict(color="black", width=0.5),
            label=d_sankey['labels'],
            color=d_sankey['node_color'],
            customdata=d_sankey['labels']
        )
        link: Dict = dict(
            source=d_sankey['source'],  # indices correspond to labels, eg A1, A2, A1, B1, ...
            target=d_sankey['target'],
            value=d_sankey['value'],
            color=d_sankey['edge_color'],
            label=d_sankey['edge_labels'],  # over-written by hover template
            customdata=d_sankey['edge_custom_data'],
            hovertemplate='<b><i>%{label}</i></b><br />Source: %{source.customdata}<br />'
                          'Target: %{target.customdata}<br />%{customdata}'
        )
        return node, link

    def _get_scatter_node_edges(self, pos):
        # edges
        edge_color_map: Dict = {True: 'grey', False: 'red'}
        edge_annotations: Dict = {}

        edge_traces = []
        for u, v, data in self.graph.edges(data=True):
            x0, y0 = pos[u]
            x1, y1 = pos[v]
            edge_annotations[data['mc'].name] = {'pos': np.mean([pos[u], pos[v]], axis=0)}
            edge_traces.append(go.Scatter(x=[x0, x1], y=[y0, y1],
                                          line=dict(width=2, color=edge_color_map[data['mc'].status.ok]),
                                          hoverinfo='none',
                                          mode='lines+markers',
                                          text=str(data['mc'].name),
                                          marker=dict(
                                              symbol="arrow",
                                              color=edge_color_map[data['mc'].status.ok],
                                              size=16,
                                              angleref="previous",
                                              standoff=15)
                                          ))

        # nodes
        node_color_map: Dict = {None: 'grey', True: 'green', False: 'red'}
        node_x = []
        node_y = []
        node_color = []
        node_text = []
        node_label = []
        for node in self.graph.nodes():
            x, y = pos[node]
            node_x.append(x)
            node_y.append(y)
            node_color.append(node_color_map[self.graph.nodes[node]['mc'].is_balanced])
            node_text.append(node)
            node_label.append(self.graph.nodes[node]['mc'].name)
        node_trace = go.Scatter(
            x=node_x, y=node_y,
            mode='markers+text',
            hoverinfo='none',
            marker=dict(
                color=node_color,
                size=30,
                line_width=2),
            text=node_text,
            customdata=node_label,
            hovertemplate='%{customdata}<extra></extra>')

        # edge annotations
        edge_labels = list(edge_annotations.keys())
        edge_label_x = [edge_annotations[k]['pos'][0] for k, v in edge_annotations.items()]
        edge_label_y = [edge_annotations[k]['pos'][1] for k, v in edge_annotations.items()]

        edge_annotation_trace = go.Scatter(
            x=edge_label_x, y=edge_label_y,
            mode='markers',
            hoverinfo='text',
            marker=dict(
                color='grey',
                size=3,
                line_width=1),
            text=edge_labels)

        return edge_traces, node_trace, edge_annotation_trace

    @staticmethod
    def _get_position_kwargs(table_pos, table_area, plot_type):
        """Helper to manage location dependencies

        Args:
            table_pos: position of the table: left|right|top|bottom
            table_area: fraction of the plot to assign to the table [0, 1]

        Returns:

        """
        name_type_map: Dict = {'sankey': 'sankey', 'network': 'xy'}
        specs = [[{"type": 'table'}, {"type": name_type_map[plot_type]}]]

        widths: Optional[List[float]] = [table_area, 1.0 - table_area]
        subplot_kwargs: Dict = {'rows': 1, 'cols': 2, 'specs': specs}
        table_kwargs: Dict = {'row': 1, 'col': 1}
        plot_kwargs: Dict = {'row': 1, 'col': 2}

        if table_pos == 'left':
            subplot_kwargs['column_widths'] = widths
        elif table_pos == 'right':
            subplot_kwargs['column_widths'] = widths[::-1]
            subplot_kwargs['specs'] = [[{"type": name_type_map[plot_type]}, {"type": 'table'}]]
            table_kwargs['col'] = 2
            plot_kwargs['col'] = 1
        else:
            subplot_kwargs['rows'] = 2
            subplot_kwargs['cols'] = 1
            table_kwargs['col'] = 1
            plot_kwargs['col'] = 1
            if table_pos == 'top':
                subplot_kwargs['row_heights'] = widths
                subplot_kwargs['specs'] = [[{"type": 'table'}], [{"type": name_type_map[plot_type]}]]
                table_kwargs['row'] = 1
                plot_kwargs['row'] = 2
            elif table_pos == 'bottom':
                subplot_kwargs['row_heights'] = widths[::-1]
                subplot_kwargs['specs'] = [[{"type": name_type_map[plot_type]}], [{"type": 'table'}]]
                table_kwargs['row'] = 2
                plot_kwargs['row'] = 1

        if plot_type == 'network':  # different arguments for different plots
            plot_kwargs = {f'{k}s': v for k, v in plot_kwargs.items()}

        return subplot_kwargs, table_kwargs, plot_kwargs

    def _rpt_to_html(self, df: pd.DataFrame) -> Dict:
        custom_data: Dict = {}
        fmts: Dict = self._get_column_formats(df.columns)
        for i, row in df.iterrows():
            str_data: str = '<br />'
            for k, v in dict(row).items():
                str_data += f"{k}: {v:{fmts[k][1:]}}<br />"
            custom_data[i] = str_data
        return custom_data

    @staticmethod
    def _color_from_float(vmin: float, vmax: float, val: float,
                          cmap: Union[ListedColormap, LinearSegmentedColormap]) -> Tuple[float, float, float]:
        if isinstance(cmap, ListedColormap):
            color_index: int = int((val - vmin) / ((vmax - vmin) / 256.0))
            color_index = min(max(0, color_index), 255)
            color_rgba = tuple(cmap.colors[color_index])
        elif isinstance(cmap, LinearSegmentedColormap):
            norm = matplotlib.colors.Normalize(vmin=vmin, vmax=vmax)
            m = cm.ScalarMappable(norm=norm, cmap=cmap)
            r, g, b, a = m.to_rgba(val, bytes=True)
            color_rgba = int(r), int(g), int(b), int(a)
        else:
            NotImplementedError("Unrecognised colormap type")

        return color_rgba

    def set_node_names(self, node_names: Dict[int, str]):
        """Set the names of network nodes with a Dict
        """
        for node in node_names.keys():
            if ('mc' in self.graph.nodes[node].keys()) and (node in node_names.keys()):
                self.graph.nodes[node]['mc'].name = node_names[node]


    def set_stream_data(self, stream_data: dict[str, Optional[MC]]):
        """Set the data (MassComposition) of network edges (streams) with a Dict"""
        for stream_name, stream_data in stream_data.items():
            stream_found = False
            nodes_to_refresh = set()
            for u, v, data in self.graph.edges(data=True):
                if 'mc' in data.keys() and (data['mc'].name if data['mc'] is not None else data['name']) == stream_name:
                    self._logger.info(f'Setting data on stream {stream_name}')
                    data['mc'] = stream_data
                    stream_found = True
                    nodes_to_refresh.update([u, v])
            if not stream_found:
                self._logger.warning(f'Stream {stream_name} not found in graph')
            else:
                # refresh the node status
                for node in nodes_to_refresh:
                    self.graph.nodes[node]['mc'].inputs = [self.graph.get_edge_data(e[0], e[1])['mc'] for e in
                                                           self.graph.in_edges(node)]
                    self.graph.nodes[node]['mc'].outputs = [self.graph.get_edge_data(e[0], e[1])['mc'] for e in
                                                            self.graph.out_edges(node)]


    def set_operation_data(self, node):
        """Set the input and output data for a node.
        Uses the data on the edges (streams) connected to the node to refresh the data and check for node balance.
        """
        node_data: Operation = self.graph.nodes[node]['mc']
        node_data.inputs = [self.graph.get_edge_data(e[0], e[1])['mc'] for e in self.graph.in_edges(node)]
        node_data.outputs = [self.graph.get_edge_data(e[0], e[1])['mc'] for e in self.graph.out_edges(node)]
        node_data.check_balance()


    def streams_to_dict(self) -> Dict[str, MC]:
        """Export the Stream objects to a Dict

        Returns:
            A dictionary keyed by name containing MassComposition objects

        """
        streams: Dict[str, MC] = {}
        for u, v, data in self.graph.edges(data=True):
            if 'mc' in data.keys():
                streams[data['mc'].name] = data['mc']
        return streams


    def nodes_to_dict(self) -> Dict[int, OP]:
        """Export the MCNode objects to a Dict

        Returns:
            A dictionary keyed by integer containing MCNode objects

        """
        nodes: Dict[int, OP] = {}
        for node in self.graph.nodes.keys():
            if 'mc' in self.graph.nodes[node].keys():
                nodes[node] = self.graph.nodes[node]['mc']
        return nodes


    def set_nodes(self, stream: str, nodes: Tuple[int, int]):
        mc: MC = self.get_stream_by_name(stream)
        mc._nodes = nodes
        self._update_graph(mc)


    def reset_nodes(self, stream: Optional[str] = None):
        """Reset stream nodes to break relationships

        Args:
            stream: The optional stream (edge) within the network.
              If None all streams nodes on the network will be reset.


        Returns:

        """
        if stream is None:
            streams: Dict[str, MC] = self.streams_to_dict()
            for k, v in streams.items():
                streams[k] = v.set_nodes([uuid.uuid4(), uuid.uuid4()])
            self.graph = Flowsheet(name=self.name).from_objects(objects=list(streams.values())).graph
        else:
            mc: MC = self.get_stream_by_name(stream)
            mc.set_nodes([uuid.uuid4(), uuid.uuid4()])
            self._update_graph(mc)


    def _update_graph(self, mc: MC):
        """Update the graph with an existing stream object

        Args:
            mc: The stream object

        Returns:

        """
        # brutal approach - rebuild from streams
        strms: List[Union[Stream, MC]] = []
        for u, v, a in self.graph.edges(data=True):
            if a.get('mc') and a['mc'].name == mc.name:
                strms.append(mc)
            else:
                strms.append(a['mc'])
        self.graph = Flowsheet(name=self.name).from_objects(objects=strms).graph


    def get_stream_by_name(self, name: str) -> MC:
        """Get the Stream object from the network by its name

        Args:
            name: The string name of the Stream object stored on an edge in the network.

        Returns:

        """

        res: Optional[Union[Stream, MC]] = None
        for u, v, a in self.graph.edges(data=True):
            if a.get('mc') and a['mc'].name == name:
                res = a['mc']

        if not res:
            raise ValueError(f"The specified name: {name} is not found on the network.")

        return res


    def set_stream_parent(self, stream: str, parent: str):
        mc: MC = self.get_stream_by_name(stream)
        mc.set_parent_node(self.get_stream_by_name(parent))
        self._update_graph(mc)


    def set_stream_child(self, stream: str, child: str):
        mc: MC = self.get_stream_by_name(stream)
        mc.set_child_node(self.get_stream_by_name(child))
        self._update_graph(mc)


    def reset_stream_nodes(self, stream: Optional[str] = None):
        """Reset stream nodes to break relationships

        Args:
            stream: The optional stream (edge) within the network.
              If None all streams nodes on the network will be reset.


        Returns:

        """
        if stream is None:
            streams: Dict[str, MC] = self.streams_to_dict()
            for k, v in streams.items():
                streams[k] = v.set_nodes([uuid.uuid4(), uuid.uuid4()])
            self.graph = Flowsheet(name=self.name).from_objects(objects=list(streams.values())).graph
        else:
            mc: MC = self.get_stream_by_name(stream)
            mc.set_nodes([uuid.uuid4(), uuid.uuid4()])
            self._update_graph(mc)



"""
REF: https://ankitbko.github.io/blog/2021/04/logging-in-python/
"""

import functools
import logging
from datetime import datetime
from typing import Union


class MyLogger:
    def __init__(self):
        logging.basicConfig(level=logging.DEBUG,
                            format=' %(asctime)s - %(levelname)s - %(message)s')

    def get_logger(self, name=None):
        return logging.getLogger(name)


def get_default_logger():
    return MyLogger().get_logger()


def log_timer(_func=None, *, my_logger: Union[MyLogger, logging.Logger] = None):
    def decorator_log(func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            logger = get_default_logger()
            try:
                if my_logger is None:
                    first_args = next(iter(args), None)  # capture first arg to check for `self`
                    logger_params = [  # does kwargs have any logger
                                        x
                                        for x in kwargs.values()
                                        if isinstance(x, logging.Logger) or isinstance(x, MyLogger)
                                    ] + [  # # does args have any logger
                                        x
                                        for x in args
                                        if isinstance(x, logging.Logger) or isinstance(x, MyLogger)
                                    ]
                    if hasattr(first_args, "__dict__"):  # is first argument `self`
                        logger_params = logger_params + [
                            x
                            for x in first_args.__dict__.values()  # does class (dict) members have any logger
                            if isinstance(x, logging.Logger)
                               or isinstance(x, MyLogger)
                        ]
                    h_logger = next(iter(logger_params), MyLogger())  # get the next/first/default logger
                else:
                    h_logger = my_logger  # logger is passed explicitly to the decorator

                if isinstance(h_logger, MyLogger):
                    logger = h_logger.get_logger(func.__name__)
                else:
                    logger = h_logger

                # args_repr = [repr(a) for a in args]
                # kwargs_repr = [f"{k}={v!r}" for k, v in kwargs.items()]
                # signature = ", ".join(args_repr + kwargs_repr)
                # logger.debug(f"function {func.__name__} called with args {signature}")

            except Exception:
                pass

            try:
                _tic = datetime.now()
                result = func(*args, **kwargs)
                _toc = datetime.now()
                logger.info(f"Elapsed time for {func.__name__}: {_toc - _tic}")
                return result
            except Exception as e:
                logger.exception(f"Exception raised in {func.__name__}. exception: {str(e)}")
                raise e

        return wrapper

    if _func is None:
        return decorator_log
    else:
        return decorator_log(_func)



"""
To provide sample data
"""
import random
from functools import partial
from pathlib import Path
from typing import Optional, Iterable, List

import numpy as np
import pandas as pd

from elphick.geomet import Sample, IntervalSample
from elphick.geomet.flowsheet import Flowsheet
from elphick.geomet.utils.components import is_compositional
from elphick.geomet.datasets import load_size_by_assay, load_iron_ore_sample_a072391, load_size_distribution, \
    load_a072391_met
from elphick.geomet.utils.partition import napier_munn, perfect


def sample_data(include_wet_mass: bool = True, include_dry_mass: bool = True,
                include_moisture: bool = False) -> pd.DataFrame:
    """Creates synthetic data for testing

    Args:
        include_wet_mass: If True, wet mass is included.
        include_dry_mass: If True, dry mass is included.
        include_moisture: If True, moisture (H2O) is included.

    Returns:

    """

    # mass_wet: pd.Series = pd.Series([100, 90, 110], name='wet_mass')
    # mass_dry: pd.Series = pd.Series([90, 80, 100], name='dry_mass')
    mass_wet: pd.Series = pd.Series([100., 90., 110.], name='wet_mass')
    mass_dry: pd.Series = pd.Series([90., 80., 90.], name='mass_dry')
    chem: pd.DataFrame = pd.DataFrame.from_dict({'FE': [57., 59., 61.],
                                                 'SIO2': [5.2, 3.1, 2.2],
                                                 'al2o3': [3.0, 1.7, 0.9],
                                                 'LOI': [5.0, 4.0, 3.0]})
    attrs: pd.Series = pd.Series(['grp_1', 'grp_1', 'grp_2'], name='group')

    mass: pd.DataFrame = pd.concat([mass_wet, mass_dry], axis='columns')
    if include_wet_mass is True and mass_dry is False:
        mass = mass_wet
    elif include_dry_mass is False and mass_dry is True:
        mass = mass_dry
    elif include_dry_mass is False and mass_dry is False:
        raise AssertionError('Arguments provided result in no mass column')

    if include_moisture is True:
        moisture: pd.DataFrame = (mass_wet - mass_dry) / mass_wet * 100
        moisture.name = 'H2O'
        res: pd.DataFrame = pd.concat([mass, moisture, chem, attrs], axis='columns')
    else:
        res: pd.DataFrame = pd.concat([mass, chem, attrs], axis='columns')

    res.index.name = 'index'

    return res


def dh_intervals(n: int = 5,
                 n_dh: int = 2,
                 analytes: Optional[Iterable[str]] = ('Fe', 'Al2O3')) -> pd.DataFrame:
    """Down-samples The drillhole data for testing

    Args:
        n: Number of samples
        n_dh: The number of drill-holes included
        analytes: the analytes to include
    Returns:

    """

    df_data: pd.DataFrame = load_iron_ore_sample_a072391()
    # df_data: pd.DataFrame = pd.read_csv('../sample_data/iron_ore_sample_data.csv', index_col='index')

    drillholes: List[str] = []
    for i in range(0, n_dh):
        drillholes.append(random.choice(list(df_data['DHID'].unique())))

    df_data = df_data.query('DHID in @drillholes').groupby('DHID').sample(5)

    cols_to_drop = [col for col in is_compositional(df_data.columns) if (col not in analytes) and (col != 'H2O')]
    df_data.drop(columns=cols_to_drop, inplace=True)

    df_data.index.name = 'index'

    return df_data


def size_by_assay() -> pd.DataFrame:
    """ Sample Size x Assay dataset
    """

    df_data: pd.DataFrame = load_size_by_assay()

    # df_data: pd.DataFrame = pd.DataFrame(data=[size_retained, size_passing, mass_pct, fe, sio2, al2o3],
    #                                      index=['size_retained', 'size_passing', 'mass_pct', 'Fe', 'SiO2', 'Al2O3']).T

    # # convert the sizes from micron to mm
    # df_data[['size_retained', 'size_passing']] = df_data[['size_retained', 'size_passing']] / 1000.0

    df_data.set_index(['size_retained', 'size_passing'], inplace=True)

    # ensure we meet the input column name requirements
    df_data.rename(columns={'mass_pct': 'mass_dry'}, inplace=True)

    return df_data


def size_by_assay_2() -> pd.DataFrame:
    """ 3 x Sample Size x Assay dataset (balanced)
    """
    mc_size: IntervalSample = IntervalSample(size_by_assay(), name='feed', moisture_in_scope=False)
    partition = partial(napier_munn, d50=0.150, ep=0.1, dim='size')
    mc_coarse, mc_fine = mc_size.split_by_partition(partition_definition=partition, name_1='coarse', name_2='fine')
    fs: Flowsheet = Flowsheet().from_streams([mc_size, mc_coarse, mc_fine])
    return fs.to_dataframe()


def size_by_assay_3() -> pd.DataFrame:
    """ 3 x Sample Size x Assay dataset (unbalanced)
    """
    mc_size: IntervalSample = IntervalSample(size_by_assay(), name='feed')
    partition = partial(napier_munn, d50=0.150, ep=0.1, dim='size')
    mc_coarse, mc_fine = mc_size.split_by_partition(partition_definition=partition, name_1='coarse', name_2='fine')
    # add error to the coarse stream to create an imbalance
    df_coarse_2 = mc_coarse.data.to_dataframe().apply(lambda x: np.random.normal(loc=x, scale=np.std(x)))
    mc_coarse_2: Sample = Sample(data=df_coarse_2, name='coarse')
    mc_coarse_2 = mc_coarse_2.set_parent_node(mc_size)
    fs_ub: Flowsheet = Flowsheet().from_streams([mc_size, mc_coarse_2, mc_fine])
    return fs_ub.to_dataframe()


def size_distribution() -> pd.DataFrame:
    return load_size_distribution()


def iron_ore_sample_data() -> pd.DataFrame:
    return load_iron_ore_sample_a072391().set_index('index')


def iron_ore_met_sample_data() -> pd.DataFrame:
    df_met: pd.DataFrame = load_a072391_met()
    df_met.dropna(subset=['Dry Weight Lump (kg)'], inplace=True)
    df_met['Dry Weight Lump (kg)'] = df_met['Dry Weight Lump (kg)'].apply(lambda x: x.replace('..', '.')).astype(
        'float64')
    df_met['Fe'] = df_met['Fe'].replace('MISSING', np.nan).astype('float64')
    df_met.dropna(subset=['Fe', 'Bulk_Hole_No', 'Dry Weight Fines (kg)'], inplace=True)
    df_met.columns = [col.replace('LOITotal', 'LOI') for col in df_met.columns]
    df_met.columns = [
        col.strip().lower().replace(' ', '_').replace('(', '').replace(')', '').replace('%', 'pct').replace('__', '_')
        for
        col in df_met.columns]

    # clean up some values and types
    df_met = df_met.replace('-', np.nan).replace('#VALUE!', np.nan)
    head_cols: List[str] = [col for col in df_met.columns if 'head' in col]
    df_met[head_cols] = df_met[head_cols].astype('float64')
    df_met['bulk_hole_no'] = df_met['bulk_hole_no'].astype('category')
    df_met['sample_number'] = df_met['sample_number'].astype('int64')
    df_met.set_index('sample_number', inplace=True)

    # moves suffixes to prefix
    df_met = df_met.pipe(_move_suffix_to_prefix, '_head')
    df_met = df_met.pipe(_move_suffix_to_prefix, '_lump')
    return df_met


def demo_size_network() -> Flowsheet:
    mc_size: Sample = Sample(size_by_assay(), name='size sample')
    partition = partial(perfect, d50=0.150, dim='size')
    mc_coarse, mc_fine = mc_size.split_by_partition(partition_definition=partition)
    mc_coarse.name = 'coarse'
    mc_fine.name = 'fine'
    fs: Flowsheet = Flowsheet().from_streams([mc_size, mc_coarse, mc_fine])
    return fs


def _move_suffix_to_prefix(df, suffix):
    suffix_length = len(suffix)
    for col in df.columns:
        if col.endswith(suffix):
            new_col = suffix[1:] + '_' + col[:-suffix_length]  # Remove the suffix and prepend it to the start
            df.rename(columns={col: new_col}, inplace=True)
    return df


if __name__ == '__main__':
    df1: pd.DataFrame = size_by_assay()
    df2: pd.DataFrame = size_by_assay_2()
    df3: pd.DataFrame = size_by_assay_3()
    df4: pd.DataFrame = iron_ore_met_sample_data()
    print('done')
